/*
 * TestResult.h
 */

#ifndef cppunit_TestResult_h
#define cppunit_TestResult_h

#include "cuconfig.h"

cu_NAMESPACE_BEGIN(cppunit)

class Test;
class TestFailure;
class TestListener;

/**
 * ƥȷ̤ݻ륯饹<br>
 * A TestResult collects the results of executing a test case.
 * It is an instance of the Collecting Parameter pattern.
 *
 * The test framework distinguishes between failures and errors.<br>
 * A failure is anticipated and checked for with assertions.<br>
 * Errors are unanticipated problems signified by exceptions
 * that are not generated by the framework.
 *
 * ͭ TestFailure Υ饤ե롣<br>
 * TestListener Υ饤եϥ饤¦Ǵ뤳ȡ
 *
 * @see Test
 */
class TestResult
{
  public:
	/** ¾ޥեδ쥯饹 */
	class Mutex {
	  public:
		virtual ~Mutex() {}
		virtual void lock();
		virtual void unlock();
	};

  protected:
	/** Mutex ưå */
	class AutoMutexLock {
	  public:
		/// å롣
		explicit AutoMutexLock(Mutex *mutex)
			: m_mutex(mutex) { m_mutex->lock(); }
		/// å롣
		~AutoMutexLock() { m_mutex->unlock(); }

	  private:
		AutoMutexLock(const AutoMutexLock& rhs);
		AutoMutexLock& operator=(const AutoMutexLock& rhs);
		Mutex* m_mutex;
	};

  public:
	/**
	 * 󥹥ȥ饯
	 * @param mutex 0 ޤ <code>new</code> 줿¾ޥե
	 */
	explicit TestResult(Mutex* mutex = 0);

	/** ǥȥ饯 */
	~TestResult();

	/**
	 * ꥹʤɲä롣
	 *
	 * ƥȼ¹ϻѤʤȡ
	 */
	void addListener(TestListener* listener);

	/**
	 * ƥȤ򳫻Ϥ롣<br>
	 * Informs the result that a test will be started.
	 */
	void startTest(Test* test);

	/**
	 * ƥȤλ<br>
	 * Informs the result that a test was completed.
	 */
	void endTest(Test* test);

	/**
	 * ƥȤμԤɲä롣<br>
	 * Adds a failure to the list of failures.
	 * @param failure ԤƥȤξ
	 */
	void addFailure(const TestFailure* failure);

	/**
	 * 顼ɲä롣<br>
	 * Adds an error to the list of errors.
	 * @param error 顼
	 */
	void addError(const TestFailure* error);

	/**
	 * ¹Ԥ줿ƥȤθĿ֤<br>
	 * Gets the number of run tests.
	 */
	int runCount() const;

	/**
	 * ԤƥȤθĿ֤<br>
	 * Gets the number of detected failures.
	 */
	int failureCount() const;

	/**
	 * 顼ˤʤäƥȤθĿ֤<br>
	 * Gets the number of detected errors.
	 */
	int errorCount() const;

	/**
	 * ϢΥƥȤ?<br>
	 * Returns whether the entire test was successful or not.
	 */
	bool wasSuccessful() const;

	/**
	 * ƥȤߤɬפ뤫?<br>
	 * Returns whether testing should be stopped.
	 */
	bool shouldStop() const;

	/**
	 * ƥȤߤ롣<br>
	 * Stop testing.
	 */
	void stop();

	/**
	 * ԤƥȤΥꥹȤ֤<br>
	 * Returns a list of the failures.
	 *
	 * ƥȼ¹ϻѤʤȡ
	 */
	const cu_List& failures() const { return m_failures; }

	/**
	 * 顼ΥꥹȤ֤<br>
	 * Returns a list of the errors.
	 *
	 * ƥȼ¹ϻѤʤȡ
	 */
	const cu_List& errors() const { return m_errors; }

  private:
	TestResult(const TestResult& rhs);
	TestResult& operator=(const TestResult& rhs);

  private:
	Mutex* m_mutex;
	cu_List m_listeners;	// list<TestListener*>
	cu_List m_failures;		// list<TestFailure*>
	cu_List m_errors;		// list<TestFailure*>
	int  m_runCount;
	bool m_shouldStop;
};

cu_NAMESPACE_END

#endif /* !cppunit_TestResult_h */
