/*
 * TestRunner.h
 */

#ifndef cppunit_TestRunner_h
#define cppunit_TestRunner_h

#include "cuconfig.h"
#include "TestRegistry.h"

cu_NAMESPACE_BEGIN(cppunit)

class Test;
class TestResult;

/**
 * ƥǥƥȥ<br>
 * A text mode test runner.
 *
 * The test runner manage the life cycle of the added tests.
 *
 * TestRunner prints out a trace as the tests are executed followed by
 * a summary at the end.
 *
 * Here is an example:
 * <pre>
 * class ATestCase : public TestCase { ... };
 *
 * int main(int argc, char* argv[]) {
 *     TestRunner runner;
 *     runner.addTest("ATestCase", new ATestCase("ATestCase"));
 *     return runner.run(argc, argv);
 * }
 * </pre>
 */
class TestRunner
{
  public:
	/** 󥹥ȥ饯 */
	TestRunner();

	/** ǥȥ饯 */
	~TestRunner() {}

	/**
	 * ƥȤɲä롣
	 * @param test ɲäƥȡ
	 */
	void addTest(Test* test);

	/**
	 * ƥȤɲä롣
	 * @param name ƥ̾
	 * @param test ɲäƥȡ
	 */
	void addTest(const char* name, Test* test);

	/**
	 * ƥȤ¹Ԥ롣
	 * @param argc ޥɥ饤θĿ
	 * @param argv ޥɥ饤
	 * @return ¹Է̡0: SUCCESS, 1: FAILURE, etc.
	 */
	int run(int argc, char* argv[]);

  protected:
	/**
	 * ޥɥ饤Ϥ롣
	 * @param argc ޥɥ饤θĿ
	 * @param argv ޥɥ饤
	 */
	int parseArgs(int argc, char* argv[]);

	/** ƥȤ¹Ԥ롣 */
	int runTests();

	/** ܡɤϤԤġ */
	void wait();

	/**
	 * ˡɽ롣
	 * @param command ޥ̾
	 */
	void printUsage(const char* command) const;

	/**
	 * ƥȷ̤ɽ롣
	 * @param result ƥȷ̡
	 */
	void printResult(const TestResult& result) const;

	/**
	 * ƥȷ̤Υإåɽ롣
	 * @param result ƥȷ̡
	 */
	void printHeader(const TestResult& result) const;

	/**
	 * ԤƥȤΥꥹȤɽ롣
	 * @param failures ԤƥȤΥꥹȡ
	 */
	void printFailures(const cu_List& failures) const;

	/**
	 * 顼ΥꥹȤɽ롣
	 * @param errors 顼Υꥹȡ
	 */
	void printErrors(const cu_List& errors) const;

  private:
	TestRunner(const TestRunner& rhs);
	TestRunner& operator=(const TestRunner& rhs);

  private:
	bool m_help;
	bool m_wait;
	bool m_verbose;
	const char*  m_logfile;
	TestRegistry m_registry;

  private:
	class Utility;
	class SimpleListener;
	class Logger;
#if defined(__sparc)
	friend class Utility;
	friend class SimpleListener;
	friend class Logger;
#endif
};

cu_NAMESPACE_END

#endif /* !cppunit_TestRunner_h */
