#ifndef HOTARU_BASIC_H_
#define HOTARU_BASIC_H_
/**
	@file hotaru_basic.h
	@brief ʽ̤Υ֥ʡ׻ġ
	@author $Author: herumi $
	@version $Revision: 1.15 $
	@date $Date: 2005/01/29 10:50:32 $
*/

#include <set>
#include <map>

namespace Hotaru {

namespace Basic {

const char* getString() { return "Basic"; }

/**
	Ĵ¿༰Υݥ󥿤ΥڥϿ
*/
struct CheckedPolyPair {
	const F2poly* f_;
	const F2poly* g_;
	/* f < g Ȥ */
	CheckedPolyPair(const F2poly* f = 0, const F2poly* g = 0)
	{
		if (f < g) {
			f_ = f, g_ = g;
		} else {
			f_ = g, g_ = f;
		}
	}
	bool operator<(const CheckedPolyPair& rhs) const
	{
		if (f_ < rhs.f_ || (f_ == rhs.f_ && g_ < rhs.g_)) {
			return true;
		}
		return false;
	}
};

#if 0
/*
	¿ʬǥڥΥꥹȤäƤ*ɤ*άǼФȤ
	ۤisGreater()ν򤤤褤Ȼפ
	Ǥʤ(=̤ξ¤Ǥʤ)ȳ껻Ȥޤ
	ʤȤ롥
*/
/*
	Τۤ褤Ȼפɻ֤ؤѤ餺
*/
class CheckedHistory {
private:
	/*
		fФf(ݥ󥿤Ȥ)礭g⤿
	*/
	typedef std::set<const F2poly*> CheckedList;
	typedef std::map<const F2poly*, CheckedList> CheckedListMap;
	CheckedListMap list_;
public:
	/*
		ڥ(f, g)Ĵݤ򸡺
	*/
	inline bool findPair(const F2poly* f, const F2poly* g) const
	{
		if (g < f) {
			std::swap(f, g);
		}
		CheckedListMap::const_iterator itr = list_.find(f);
		if (itr != list_.end()) {
			return itr->second.find(g) != itr->second.end();
		} else {
			return false;
		}
	}

	/*
		ڥ(f, g)ĴȤˤ
	*/
	inline void set(const F2poly* f, const F2poly* g)
	{
		if (g < f) {
			std::swap(f, g);
		}
		CheckedListMap::iterator itr = list_.find(f);
		if (itr != list_.end()) {
			itr->second.insert(g);
		} else {
			/* fä */
			CheckedList entry;
			entry.insert(g);
			list_[f] = entry;
		}
	}
};
#else
class CheckedHistory {
private:
	std::set<CheckedPolyPair> checkedPairs_;
public:
	inline bool findPair(const F2poly* f, const F2poly* g) const
	{
		return checkedPairs_.find(CheckedPolyPair(f, g)) != checkedPairs_.end();
	}

	inline void set(const F2poly* f, const F2poly* g)
	{
		checkedPairs_.insert(CheckedPolyPair(f, g));
	}
};
#endif

/**
	G = {g1, ..., gm}֥ʡ줫ݤ
	@note Gϴޤޤʤ
	@param G [in] Ƚꤷ¿༰
	@param s [out] Ƚ˼ԤS¿༰
	@note X^2 + Xηۤ˴ޤޤƤΤǡ
	ȤSpolyˤĤƤͤɬפ롥
	LM(f)x_iޤȤS(f, x_i^2+x_i)=x_i fȤʤ롥
*/
static bool IsGbase(const F2poly::Set& G, F2poly& s, CheckedHistory& history)
{
	F2poly::Set::const_iterator i, j, prevEndG = G.end();
	--prevEndG;

	/**
		X^2 + XȤSpolyͤ
		f2poly.hisGreater()numTermӤ򤹤
		˸ƤӽФä®ʤ
	*/
	for (i = G.begin(); i != G.end(); ++i) {

		const F2term& LMs = i->LM();

		for (size_t j = 0; j < Hotaru::MAX_NUM_OF_VARIABLE; j++) {
			if (LMs.hasVar(j)) {
				s = F2poly::GetSpoly(*i, j);
				s.dividedBy(G);
				if (!s.isZero()) {
					return false;
				}
			}
		}
	}

	assert(G.size() >= 2);

	for (i = G.begin(); i != prevEndG; ++i) { // i = 0; i < m - 1; i++
		assert(!i->isZero());
//		const F2term& LMi = i->LM();

		for (j = i, ++j; j != G.end(); ++j) { // j = i + 1, j < m; j++
			assert(!j->isZero());

			if (!history.findPair(&*i, &*j)) {
//std::cout << "f=" << *i << ", g=" << *j;
//printf("(%p, %p)\n", &*i, &*j);
				history.set(&*i, &*j);
				/*
					Lemmma LCM(f, g) = 1ʤS(f, g){f, g}ˤĤƤ;0
					ʤΤǥåפǤ(ΤȤۤξ̤ʤ)
				*/
//				if (j->LM().isPrimeTo(LMi)) {
//					continue;
//				}
				s = F2poly::GetSpoly(*i, *j);
				s.dividedBy(G);
				if (!s.isZero()) {
					return false;
				}
			} else {
				/* none */
			}
		}
	}
	return true;
}

/*
	ξ¤ؤ
	0ƱΤϺ
*/
static void sortAndRemoveTrivialPoly(F2poly::Set& G)
{
	G.sort(); /* ξ¤ؤ */
	G.unique(); /* ƱΤϺ */

	/* 0 */
	F2poly::Set::iterator itr;
	itr = G.begin();
	while (itr != G.end()) {
		if (itr->isZero()) {
			itr = G.erase(itr);
		} else {
			++itr;
		}
	}
	/* 1ޤtrivial */
	if (G.size() > 0 && G.begin()->getMltDeg() == 0) {
		G.clear();
		G.push_back(F2poly(F2term())); /* = 1 */
	}
}

/**
	ľʹ
	@param G [in] ֥ʡ
	@note LMߤפʹ
*/
static void Simplify(F2poly::Set& G)
{
	sortAndRemoveTrivialPoly(G);

	/* ब2İʾʤнλ */
	if (G.size() < 2) {
		return;
	}

	/*
		˳äƤߤ
	*/
	F2poly::Set::iterator i, j, prevEndG = G.end();
	--prevEndG;

	for (i = G.begin(); i != prevEndG; ++i) { // i = 0; i < m - 1; i++
		if (i->isZero()) {
			continue;
		}
		const F2term& LM = i->LM();

		for (j = i, ++j; j != G.end(); ++j) { // j = i + 1, j < m; j++
			if (!j->isZero() && j->LM().canBeDividedBy(LM)) {
				j->dividedBy(*i);
			}
		}
	}

	sortAndRemoveTrivialPoly(G);
}

/**
	󥰥֥ʡ(դ¸ߤ)
	@param G[in] ֥ʡ
	G[j] ˸ñ༰ LM(G[i])ǳʤ for i != j
*/
static void Reduce(F2poly::Set& G)
{
RETRY:
	sortAndRemoveTrivialPoly(G);

	/*
		ब2İʾʤнλ
	*/
	if (G.size() < 2) {
		return;
	}

	bool isChanged = false;

	/*
		˳äƤߤ
	*/
	F2poly::Set::iterator i, j, prevEndG = G.end();
	--prevEndG;

	for (i = G.begin(); i != prevEndG; ++i) { // i = 0; i < m - 1; i++
		if (i->isZero()) {
			continue;
		}

		for (j = i, ++j; j != G.end(); ++j) { // j = i + 1, j < m; j++
			if (!j->isZero()) {
				isChanged |= j->reducedBy(*i);
			}
		}
	}

	/* QQQ:ȴ(äȱ黻̤ϸ餻) */
	if (isChanged) {
		goto RETRY;
	}
}

/**
	GB
	@param G [inout] in:¿༰ out:󥰥֥ʡ
*/
static void MakeGbaseBy(F2poly::Set& G)
{
	F2poly s;

/*
	Recude()ʤΤܼǤϤʤ
	cyclic(n)η׻Ū(100)᤯뤿θ©ʼ
	Reduce()ФƤƱͤ®٤Ф褦ͻ٤
*/
//	Reduce(G);

//PutF2polySet(G, 'T');

	/*
		SpolyȤäƤߤ
		50=>47@mingw
	*/
	CheckedHistory history;
	/*
		
		IsGbase()Ƥ֤ȤG0ޤޤʤ
	*/
	int count = 0;
	while (!IsGbase(G, s, history)) {
		if (debug > 1) std::cout << "add poly:" << s << std::endl;
		G.push_back(s);
		/*
			ʤ٤ñפʹˡͤ뤳
			ΤȤReduceϽŤΤǻƤӽФʤ
			GGBǤȤIsGbase()ƤReduce()
			ƤǤȹ®Ǥ
			ߤŬ˴֤򤪤ƸƤ֤Ȥˤ
		*/
		if ((count & 7) == 0) {
			Simplify(G);
		} else {
			Reduce(G);
		}
		count++;
printf("num of G=%d\n", (int)G.size());
	}
puts("loop end\nnow reducing");

	/*
		while()Reduce()򤷤Ƥ¤פä̤
		򤷤Ȥϥ롼פȴȤReduce()ƤȤϸ¤ʤ
	*/
	Reduce(G);

	/*
		ReduceƤ⥰֥ʡǤ뤳Ȥ¸ȻפΤ
		ǤϤʤʤ? => 2004/11/24  X^2+XηθƤʤä
	*/
	assert(IsGbase(G, s, history));
}

} // end of namespace Basic

} // end of namespace Hotaru

#endif /* HOTARU_BASIC_H_ */
