#ifndef   HOTARU_F4_H_
#define   HOTARU_F4_H_

/**
	@file hotaru_f4.h
	@brief Groebner Base(F4)
	@author $Author: dskw $
	@version $Revision: 1.54 $
	@date $Date: 2005/02/26 07:41:51 $
*/


#include "hotaru_f4_poly.h"
#include <time.h>
#include <set>
#include <iterator>
//#include <map>

// ======================================================================
//  container setting
// ======================================================================

#if defined(USE_MYLIST) || defined(USE_MYSET)
	#ifndef USE_MYLIST 
		#include "mylist.h"
	#endif
	#define PAIR_CONT mylist
	#define PAIR_OPT_ALLOC
#else
	#define PAIR_CONT std::list
	#ifdef USE_MYALLOC
		#define PAIR_OPT_ALLOC , myalloc<Pair> /**/
	#elif defiend(USE_BOOST)
		#define PAIR_OPT_ALLOC , boost::pool_allocator<Pair> /**/
	#else
		#define PAIR_OPT_ALLOC
	#endif
#endif


// ======================================================================
//  implementation
// ======================================================================

namespace Hotaru {
namespace F4 {

// ----------------------------------------------------------------------
//  external symbols
// ----------------------------------------------------------------------
#ifndef USE_F5
const char* getString() { return "F4"; }
#else
const char* getString() { return "F5"; }
#endif
void MakeGbaseBy(F2poly::Set& G);



// ----------------------------------------------------------------------
//  debug settings
// ----------------------------------------------------------------------
#ifndef PROF
	#ifdef USE_PARAF4
		#define PROF 1
	#else
		#define PROF 2
	#endif
#endif
#if(PROF == 1)
	#define Prof(x)\
	std::cout << x << "\t\t#G = " << G.size() << " \t\r"; std::cout.flush()
	#define ProfNG(x)\
	std::cout << x << "\t\t\t\r"; std::cout.flush()
#elif(PROF == 2)
	#define Prof(x)\
	std::cout << x << "\t\t#G = " << G.size() << " \t" << std::endl
	#define ProfNG(x)\
	std::cout << x << "\t\t\t" << std::endl
#else
	#define Prof(x)
	#define ProfNG(x)
#endif

void PUTG(F2poly::Set const& G){
std::cout << "Polys[" << std::endl;
PutF2polySet(G, 'G');
std::cout << "]" << std::endl;
}

// ----------------------------------------------------------------------
//  common types
// ----------------------------------------------------------------------
typedef std::list<F2poly::Set> List;
struct Pair {
	F2poly const& f1_;
	F2poly const& f2_;
	int d_;
	int var_;
	
	int degreeLCM() const { return d_; }
	int degreeLCM_TD(int td) const {
		return (f1_.LM().getDegree() > td ||
			 f2_.LM().getDegree() > td )? 10000: d_;
	}

	int numTerms() const {
		return f1_.getNumOfTerm() +((&f1_ == &f2_)?  + 2:
			f2_.getNumOfTerm());
	}
	
	/* f, g */
	Pair(F2poly const& f1, F2poly const& f2) : f1_(f1), f2_(f2) {
		d_ = F2term::getLCM(f1_.LM(), f2_.LM()).getDegree();
	}
	
	/* f, Si */
	Pair(F2poly const& f1, int var = -1) : f1_(f1), f2_(f1), var_(var) {
		d_ = f1_.LM().getDegree() + 1;
	}
	
	struct LessDegLCM { bool operator()(Pair const& l, Pair const& r)const {
		return l.degreeLCM() < r.degreeLCM();
	}};
	
	struct LessDegLCM_TD { 
		int td_;
		LessDegLCM_TD(int td) : td_(td) {}
		bool operator()(Pair const& l, Pair const& r)const {
			return l.degreeLCM_TD(td_) < r.degreeLCM_TD(td_);
		}
	};
	
	struct LessNTerms { bool operator()(Pair const& l, Pair const& r)const {
		return l.numTerms() < r.numTerms();
	}};
	
#ifdef USE_SUGAR
	int sugar() const {
		if(&f1_ == &f2_) return 1 + f1_.sugar_;
		int d = F2term::getLCM(f1_.LM(), f2_.LM()).getDegree();
		return std::max(f1_.sugar_ - f1_.LM().getDegree(),
			f2_.sugar_ - f2_.LM().getDegree()) + d;
	}
	
	struct LessSugar { bool operator()(Pair const& l, Pair const& r)const {
		return l.sugar() < r.sugar();
	}};
#endif
	
	F2poly getSpoly() const {
		if(&f1_ == &f2_) {
			F2poly f(f1_);
			f.multiplyByX(var_);
#ifdef USE_SUGAR
			f.sugar_ = f1_.sugar_ + 1;
#endif
			return f;
		} else {
#ifdef USE_SUGAR
			F2poly f = F2poly::GetSpoly(f1_, f2_);
			f.sugar_ = std::max(f1_.sugar_ - f1_.LM().getDegree(),
				f2_.sugar_ - f2_.LM().getDegree()) +
				F2term::getLCM(f1_.LM(), f2_.LM()).getDegree();
#else
			return F2poly::GetSpoly(f1_, f2_);
#endif
		}
	}
	
void PUT() const {
std::cout << "Pair[f1 = " << f1_;
if(&f1_ != &f2_) std::cout << ", f2 = " << f2_;
else std::cout << ", var = " << var_;
std::cout << "]" << std::endl;
}

};

//typedef PAIR_CONT<Pair PAIR_OPT_ALLOC, POOL_PAIR_NUM> PairList;
struct PairList : PAIR_CONT<Pair PAIR_OPT_ALLOC, POOL_PAIR_NUM> {
void PUT() const {
std::cout << "PairList[" << std::endl;
for(const_iterator i = begin(); i != end(); ++i) i->PUT();
std::cout << "]" << std::endl;
}
};


// ----------------------------------------------------------------------
//  select functions
// ----------------------------------------------------------------------

//
// select LCM's total degree is minimum
//
#ifdef USE_HFE
int tdeg_ = 0;
#endif
struct SelectMinDegLCM { void operator()(PairList& P, PairList& Pd)const {
#ifdef USE_HFE
	int td = tdeg_;
	int d = std::min_element(P.begin(), P.end(),
			Pair::LessDegLCM_TD(td))->degreeLCM_TD(td);
std::cout << "tdeg = " << td << ", deg = " << d << std::endl;
	while(d > 9999) {
		++td;
		d = std::min_element(P.begin(), P.end(),
			Pair::LessDegLCM_TD(td))->degreeLCM_TD(td);
std::cout << "tdeg = " << td << ", deg = " << d << std::endl;
	}
#else
	int d = std::min_element(P.begin(), P.end(),
		Pair::LessDegLCM())->degreeLCM();
#endif

	PairList::iterator i = P.begin();
	while(i != P.end()) {
#ifdef USE_HFE
		if(i->degreeLCM_TD(td) == d) Pd.splice(Pd.end(), P, i++); else ++i;
#else
		if(i->degreeLCM() == d) Pd.splice(Pd.end(), P, i++); else ++i;
#endif
	}
#ifdef USE_HFEx
	if(d == 4) {
#if 0
		static int nn = 0;
		static int a[] = {2, 1, 1}, b[] = {7, 2, 3};
		int num = Pd.size()/b[nn]*a[nn]; nn = (nn+1)%3;
		for(int ii = 0; ii < num; ++ii) Pd.erase(Pd.begin());
#else
		static int nn = 0;
		if(nn == 0) ++nn;
		else {
			for(PairList::iterator ii = Pd.begin(); ii!=Pd.end();){
				if(ii->f1_.LM().hasVar(8) ||
				   ii->f2_.LM().hasVar(8)) Pd.erase(ii++);
				else ++ii;
			}
		}
#endif
	}
#endif
#ifdef USE_HFE
	if(d == 5) {
		/*for(PairList::iterator ii = Pd.begin(); ii!=Pd.end();){
			if(ii->f1_.LM().hasVar(8) ||
			   ii->f2_.LM().hasVar(8)
			) ++ii;
			else Pd.erase(ii++);
		}*/
		//Pd.sort(Pair::LessNTerms());
		//while(Pd.size() > 740) Pd.erase(Pd.begin());
		/*while(Pd.size() > 750) {
			PairList::iterator ii = Pd.end();
			Pd.erase(--ii);
		}*/
		//PairList::iterator ii = Pd.begin();
		//for(int n = 0; ii != Pd.end(); ++n) {
		//	if(n%7) Pd.erase(ii++); else ++ii;
		//}
	}
#endif
}};


//
// select specified degree
//
struct SelectDeg {
	int deg_;
	SelectDeg(int d) : deg_(d) {}
	void operator()(PairList& P, PairList& Pd)const {
		PairList::iterator i = P.begin();
		while(i != P.end()) {
			if(i->degreeLCM() == deg_) Pd.splice(Pd.end(), P, i++);
			else ++i;
		}
	}
};


#ifdef USE_SUGAR
//
// select sugar is minimum
//
static int sugar_;
struct SelectMinSugar { void operator()(PairList& P, PairList& Pd)const {
	sugar_ = std::min_element(P.begin(), P.end(),
		 Pair::LessSugar())->sugar();

	PairList::iterator i = P.begin();
	while(i != P.end()) {
		if(i->sugar() == sugar_) Pd.splice(Pd.end(), P, i++); else ++i;
	}
}};
#endif

//
// select LCM's total degree is minimum (with maximum pairs)
//
struct SelectMinDegLCM_LT {
	int maxPairs_;
	SelectMinDegLCM_LT(int m) : maxPairs_(m) {}
	
	void operator()(PairList& P, PairList& Pd)const {
		int d = std::min_element(P.begin(), P.end(),
			 Pair::LessDegLCM())->degreeLCM();
	
		int np = 0;
		PairList::iterator i = P.begin();
		while(i != P.end()) {
			if(i->degreeLCM() == d) {
				Pd.splice(Pd.end(), P, i++);
				if(++np >= maxPairs_) break;
			} else ++i;
		}
	}
};

//
// select sum of number of terms <= minimum + 5
//
struct SelectMinNTerms { void operator()(PairList& P, PairList& Pd)const {
	int n = std::min_element(P.begin(), P.end(),
		 Pair::LessNTerms())->numTerms();
	//int d = std::min_element(P.begin(), P.end(),
	//	 Pair::LessDegLCM())->degreeLCM();

	PairList::iterator i = P.begin();
	while(i != P.end()) {
		if(i->numTerms() > n + 5) {
			++i;
			continue;
		}
		Pd.splice(Pd.end(), P, i++);
	}
}};


#if 1

// ----------------------------------------------------------------------
//  update fuctions
// ----------------------------------------------------------------------
struct UpdateSimple{ void operator()(
	F2poly::Set& G, PairList& P, F2poly::Set& F)const {
	
	Reduce(F);
	if(!F.size()) return;
	
	PolySet H;
	H.addSet(G);
	if(H.reduceBy(F)) {
		//G.putOld();
		H.getSet(G);
		G.sort(lessLMnumTerms());
	}
	
	F2poly::Set::iterator iF = F.begin();
	for(; iF != F.end(); ++iF) {
		G.push_back(*iF);
		F2poly::Set::iterator ig = G.end(); --ig;

		// h, {Si}
		for(size_t i = 0; i < MAX_NUM_OF_VARIABLE; ++i) {
			if(ig->LM().hasVar(i)) P.push_back(Pair(*ig, i));
		}
	
		// h, G
		F2poly::Set::iterator i = G.begin(), j;
		for(; i != ig; ++i) {
			if(i->isZero()) continue;
		
			// Buchberger criterion1
			if(F2term::isPrimeTo(ig->LM(), i->LM())) continue;
			
			// Buchberger criterion2
			F2term lcm = F2term::getLCM(ig->LM(), i->LM());
			for(j = G.begin(); j != i; ++j) {
				if(j->isZero()) continue;
				if(lcm.canBeDividedBy(j->LM())) break;
			}
			if(j != i) continue;
			P.push_back(Pair(*ig, *i));
		}
	}
	
}};

struct UpdateGM{ void operator()(
	F2poly::Set& G, PairList& P, F2poly::Set& F)const {
	
	//Reduce(F);
	if(!F.size()) return;
#ifdef USE_SUGAR
	for(F2poly::Set::iterator iF = F.begin(); iF != F.end(); ++iF) {
		iF->sugar_ = sugar_;
	}
#endif
	
	F2poly::Set::iterator iF = F.begin(), i, j, k;
	for(; iF != F.end(); ++iF) {
		G.push_back(*iF);
		j = G.end(); --j;
		
		// consider B
		for(PairList::iterator iP = P.begin(); iP != P.end();) {
			if(&iP->f1_ == &iP->f2_) {
				if(!iP->f1_.LM().canBeDividedBy(j->LM())) {
					++iP; continue;
				}
				F2term tj = j->LM(); tj.setVar(iP->var_);
				if(iP->f1_.LM() == tj) {
					++iP; continue;
				}
			} else {
				F2term t12 = F2term::getLCM(
					iP->f1_.LM(), iP->f2_.LM());
				if(!t12.canBeDividedBy(j->LM())) {
					++iP; continue;
				}
				F2term t1j = F2term::getLCM(
					j->LM(), iP->f1_.LM());
				if(t1j == t12) { ++iP; continue; }
				F2term t2j = F2term::getLCM(
					j->LM(), iP->f2_.LM());
				if(t2j == t12) { ++iP; continue; }
			}
			P.erase(iP++);
		}
		
		// j, {Si}
#ifdef USE_HFE
		if(j->LM().getDegree() < 4)
#endif
		for(size_t ii = 0; ii < MAX_NUM_OF_VARIABLE; ++ii) {
			// Buchberger criterion1
			if(!j->LM().hasVar(ii)) continue;
			
			// consider F -- always true
			// consider M
			for(k = G.begin(); k != j; ++k) {
				if(k->LM().canBeDividedBy(j->LM())) break;
			}
			if(k != j) continue;
			
			P.push_back(Pair(*j, ii));
		}
		
		// j, G
		for(i = G.begin(); i != j; ++i) {
			// Buchverger criterion1
			if(F2term::isPrimeTo(i->LM(), j->LM())) continue;
			
			F2term tij = F2term::getLCM(i->LM(), j->LM());
#ifdef USE_HFE
			if(tij.getDegree() > 4) continue;
#endif
			
			// consider F & M
			for(k = G.begin(); k != i; ++k) {
				if(tij.canBeDividedBy(k->LM())) break;
			}
			if(k != i) continue;
			// consider M
			for(++k; k != j; ++k) {
				F2term tjk = F2term::getLCM(j->LM(), k->LM());
				if(!tij.canBeDividedBy(tjk)) continue;
				if(tij != tjk) break;
			}
			if(k != j) continue;
			
			P.push_back(Pair(*i, *j));
		}
	}
}};
	
// ----------------------------------------------------------------------
//  Reduction functions
// ----------------------------------------------------------------------
struct Mul {
	F2term  t_;
	F2poly const& f_;
	Mul(F2term t, F2poly const& f) : t_(t), f_(f) {}
	bool operator<(Mul const& rhs) const {
		int cmp = t_.isGreater(rhs.t_);
		if(cmp) return cmp < 0;
		return (long)&f_ < (long)&rhs.f_;
	}
	F2poly getMulTopHigh() const {
		F2poly f(f_);
		return f.mulTopHigh(t_);
	}
};

// assume FP:null set
struct ReductionF4 {
	void operator()(PairList const& Pd, F2poly::Set const& G,
		F2poly::Set& FP) {
		PolySet F;
#if 1
		for(PairList::const_iterator iP = Pd.begin(); iP != Pd.end();
			 ++iP) F.addPoly(iP->getSpoly());
		F.reduceBy(G);
		//F.reduceBy(G.old_);
#else
		std::set<Mul> mul;
		for(PairList::const_iterator iP = Pd.begin(); iP != Pd.end();
			 ++iP) {
			if(iP->f1_ != iP->f2_) {
				F2term lcm = F2term::getLCM(iP->f1_.LM(),
					iP->f2_.LM());
				mul.insert(Mul(lcm/iP->f1_.LM(), iP->f1_));
				mul.insert(Mul(lcm/iP->f2_.LM(), iP->f2_));
			} else {
				F2term t; t.setVar(iP->var_);
				mul.insert(Mul(t, iP->f1_));
			}
		}
		for(std::set<Mul>::iterator iM = mul.begin(); iM != mul.end();
			++iM) F.addPoly(iM->getMulTopHigh());
		// @beginReduce_
		F.reduceBy(G, false);
#endif
		F.getSet(FP);
	}
};

// ----------------------------------------------------------------------
//  Main Schemes of F4
// ----------------------------------------------------------------------
void ReduceSimple(F2poly::Set& F);

template<typename U_, typename R_>
bool CalcEnd(U_ Update, R_ Reduction, F2poly::Set& G, F2poly::Set& FP) {
	F2poly::Set G1(G);
	G1.insert(G1.end(), FP.begin(), FP.end());
	ReduceGbase(G1);
	F2poly::Set G2;
	PairList Px;
	Update(G2, Px, G1);
	G1.clear();
	Reduction(Px, G2, G1);
	if(!G1.empty()) return false;
	PolySet H;
	H.addSet(G);
	H.reduceBy(G2);
	H.getSet(G1);
	if(!G1.empty()) return false;
std::cout << "gbase!" << std::endl;
	G = G2;
	return true;
}

#ifdef USE_HFE
void
FindLinear(F2poly::Set&FP, F2poly::Set& H) {
	//Resolve(FP);
	F2poly::Set::iterator iF = FP.begin();
	while(iF != FP.end()) {
		if(iF->isZero() || iF->LM().getDegree() > 1) {
			++iF;
			continue;
		}
std::cout << "Linear found!!" << std::endl;
std::cout << "L[] = " << *iF << std::endl;
		H.splice(H.end(), FP, iF++);
	}
}

template<typename S_, typename U_, typename R_>
void AlgorithmF4(F2poly::Set& F, 
	S_ Select, U_ Update, R_ Reduction, F2poly::Set& G) {
	
	PairList P, Pd;
clock_t b;
	F2poly::Set H;
	FindLinear(F, H);
	if(H.size()) goto Linear;
	
	// initial pairs
	Update(G, P, F);
	
ProfNG("Step1");
	// step1: diff of degree-2 LM-identical pairs
	SelectDeg(2)(P, Pd);
	if(Pd.size()) {
		F2poly::Set FP;
b = clock();
		Reduction(Pd, G, FP);
b = clock() - b;
ProfNG("new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#define HFE_REDUCE
#ifdef HFE_REDUCE
b = clock();
		Reduce(FP);
b = clock() - b;
ProfNG("reduced new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#endif
		Update(G, P, FP);
	}
	
ProfNG("Step2");
	// step2: degree-2 & field + 2 degree-2 polys
	SelectDeg(3)(P, Pd);
	if(Pd.size()) {
		F2poly::Set FP;
b = clock();
		Reduction(Pd, G, FP);
b = clock() - b;
ProfNG("new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#ifdef HFE_REDUCE
b = clock();
		Reduce(FP);
b = clock() - b;
ProfNG("reduced new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#endif
		Update(G, P, FP);
	}
	
ProfNG("Step3");
	// step3: degree-3 & field + degree-(3,2)or(3,3) polys
	SelectDeg(4)(P, Pd);
	if(Pd.size()) {
		F2poly::Set FP;
b = clock();
		Reduction(Pd, G, FP);
b = clock() - b;
ProfNG("new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#ifdef HFE_REDUCE
b = clock();
		Reduce(FP);
b = clock() - b;
ProfNG("reduced new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#endif
		FindLinear(FP, H);
		if(H.size()) goto Linear;
		Update(G, P, FP);
	}
	
ProfNG("Step4");
	// step4: degree-3 & field + degree-(3,2)or(3,3) polys
	SelectDeg(4)(P, Pd);
	if(Pd.size()) {
		F2poly::Set FP;
b = clock();
		Reduction(Pd, G, FP);
b = clock() - b;
ProfNG("new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#ifdef HFE_REDUCE
b = clock();
		ReduceS(FP);
b = clock() - b;
ProfNG("reduced new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#endif
		FindLinear(FP, H);
		if(H.size()) goto Linear;
		Update(G, P, FP);
	}
	
ProfNG("Step5");
	// step5: degree-3 & field + degree-(3,2)or(3,3) polys
	SelectDeg(4)(P, Pd);
	if(Pd.size()) {
		F2poly::Set FP;
b = clock();
		Reduction(Pd, G, FP);
b = clock() - b;
ProfNG("new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#ifdef HFE_REDUCE
b = clock();
		ReduceS(FP);
b = clock() - b;
ProfNG("reduced new poly#: " << FP.size());
ProfNG("time: " << (double)b/CLOCKS_PER_SEC << "sec");
#endif
		FindLinear(FP, H);
		if(H.size()) goto Linear;
		Update(G, P, FP);
	}
	
	Linear:;
	if(H.size()) {
		//remake problem
		Reduce(H);
		G.clear();
		P.clear();
		Pd.clear();
		PolySet X;
		X.addSet(F);
		F.clear();
		X.reduceBy(H);
		X.getSet(F);
		// recurse
		AlgorithmF4(F, Select, Update, Reduction, G);
		G.splice(G.end(), H, H.begin(), H.end());
	}
}
	
	
#else
// assume G:null set
template<typename S_, typename U_, typename R_>
void AlgorithmF4(F2poly::Set& F, 
	S_ Select, U_ Update, R_ Reduction, F2poly::Set& G) {
	PairList P;
	
	Update(G, P, F);
	
//	int wc = 0;
	while(P.size() > 0) {
//ProfNG("\n loop start:\n");
//PUTG(G); P.PUT();
		PairList Pd;
		Select(P, Pd);
		if(P.size() < 10) while(P.size()) {
			Pd.splice(Pd.end(), P, P.begin());
		}
ProfNG("Selected#: " << Pd.size() << " / " << P.size()+Pd.size());
//ProfNG("\n selected :\n");
//Pd.PUT();
		F2poly::Set FP;
		Reduction(Pd, G, FP);
#if 0		
		if(++wc == 3) {
ProfNG("reducing....");
			P.clear();
			PolySet H;
			H.addSet(G);
			G.clear();
			H.reduceBy(FP);
			H.getSet(FP);
		}
#endif
		
		Update(G, P, FP);
		
ProfNG("\n #G = " << G.size() << ", #P = " << P.size() << "\n");
	}
//ProfNG("\n result \n");
//PUTG(G);
}
#endif

template<typename S_, typename U_, typename R_>
void AlgorithmIncrementalF5(F2poly::Set& F, 
	S_ Select, U_ Update, R_ Reduction, F2poly::Set& G) {
	F2poly::Set::const_iterator ii;
	for(ii = F.begin(); ii != F.end(); ++ii) {
		F2poly::Set X; X.push_back(*ii);
		PairList P;
		Update(G, P, X);
		while(P.size() > 0) {
			PairList Pd;
			Select(P, Pd);
			F2poly::Set FP;
			Reduction(Pd, G, FP);
			Update(G, P, FP);
		}
		Reduce(G);
	}
}

// ----------------------------------------------------------------------
//  tmp
// ----------------------------------------------------------------------
void
ReduceSimple(F2poly::Set& F) {
	bool changed;
	F2poly::Set::iterator i, j;
	do {
		changed = false;
		F.sort();
		for(i = F.begin(); i != F.end(); ) {
			if(i->isZero()) {
				F.erase(i++);
				continue;
			}
			for(j = F.begin(); j != F.end(); ++j) {
				if(j == i) continue;
				if(j->isZero()) continue;
				changed |= j->reducedBy(*i);
			}
			++i;
		}
	} while(changed);
}

struct CompNumTerms { bool operator()(F2poly const& l, F2poly const& r) {
	return l.getNumOfTerm() < r.getNumOfTerm();
}};
void
Resolve(F2poly::Set& F) {
	F2poly::Set G;
	F.sort(CompNumTerms());
	F2poly::Set::iterator iS = F.begin();
	for(; iS != F.end(); ++iS) {
		if(iS->getNumOfTerm() < 2) continue;
		if(iS->getNumOfTerm() == 2) {
			F2poly::terms k = iS->first(); ++k;
			if(!iS->LM().canBeDividedBy(*k)) continue;
			F2term x = iS->LM()/(*k);
			if(x.getDegree() < 2) continue;
			for(size_t i = 0; i < MAX_NUM_OF_VARIABLE; ++i) {
				if(x.hasVar(i)) {
					F2term y(*k); y.setVar(i);
					G.push_back(F2poly(*k));
					G.back() += y;
				}
			}
			continue;
		}
		for(size_t i = 0; i < MAX_NUM_OF_VARIABLE; ++i) {
			F2term const* px = 0;
			int cn = 0;
			F2poly::terms k = iS->last(); --k;
			for(; k != iS->first(); --k) if(!k->hasVar(i)) {
				if(++cn > 1) break;
				px = &*k;
			}
			if(cn > 1) continue;
			if(!k->hasVar(i)) {
				if(++cn > 1) continue;
				px = &*k;
			}
			if(cn < 1) continue;
			F2term t(*px);
			G.push_back(F2poly(t));
			t.setVar(i); G.back() += t;
		}
	}
	F.splice(F.end(), G, G.begin(), G.end());
}


// ----------------------------------------------------------------------
//  instanciation
// ----------------------------------------------------------------------
void MakeGbaseBy(F2poly::Set& G) {
	F2poly::Set F;
#ifndef USE_HFE
	Resolve(G);
	Reduce(G);
#endif
#ifdef USE_SUGAR
	for(F2poly::Set::iterator iG = G.begin(); iG != G.end(); ++iG) {
		iG->sugar_ = iG->LM().getDegree();
	}
#endif
	
	//ReductionWithSimplify R;
	ReductionF4 R;
#ifdef USE_F5
	AlgorithmIncrementalF5(G, 
#else
	AlgorithmF4(G, 
#endif
#ifdef USE_SUGAR
		SelectMinSugar(),
#else
		//SelectMinDegLCM_LT(512),
		SelectMinDegLCM(),
#endif
		//SelectMinNTerms(),
		//UpdateSimple(),
		UpdateGM(),
		//UpdateF5(),
		R, F);
//ProfNG("en = " << R.nEnter_ << 
//	", re = " << R.nReenter_ << ", ma = " << R.nMatch_);
	G.clear();
	G.splice(G.end(), F, F.begin(), F.end());
	//ReduceSimple(G);
	ReduceGbase(G);
ProfNG("\n");
}

#else

// ----------------------------------------------------------------------
//  update fuctions
// ----------------------------------------------------------------------

//
// update with F5 criterion
//

struct UpdateF5 { void operator()(
	F2poly::Set& G, PairList& P, F2poly const& h)const {
	// @@unimplement
}};


//
// update with Buchberger criteria
//

struct UpdateSimple { void operator()(
	F2poly::Set& G, PairList& P, F2poly const& h)const {
//Prof("UpdateSimple Enter");
	G.push_back(h);
	F2poly::Set::iterator ig = G.end(); --ig;
	
	// h, {Si}
	P.push_back(Pair(*ig));
	
	// h, G
	F2poly::Set::iterator i = G.begin(), j;
	for(; i != ig; ++i) {
		if(i->isZero()) continue;
		
		// Buchberger criterion
		if(F2term::isPrimeTo(ig->LM(), i->LM())) continue;
		
		// Buchberger criterion2
#define BUCHBERGER2
#ifdef BUCHBERGER2
		F2term lcm = F2term::getLCM(ig->LM(), i->LM());
		for(j = G.begin(); j != i; ++j) {
			if(j->isZero()) continue;
			if(lcm.canBeDividedBy(j->LM())) break;
		}
		if(j != i) continue;
#endif
		
		P.push_back(Pair(*ig, *i));
	}
//Prof("UpdateSimple Exit");
}};


// ----------------------------------------------------------------------
//  matrix operations
// ----------------------------------------------------------------------
void
ReduceSimple(F2poly::Set& F);

void
MinimumRowEchelonForm(F2poly::Set const& F, F2poly const& Terms,
	F2poly::Set& FT) {
	
	F2poly::Set::const_iterator i;
	size_t nRows = 0;
	for(i = F.begin(); i != F.end(); ++i) if(!i->isZero()) ++nRows;
	if(nRows < 1) return;
	
	size_t nCols = Terms.getNumOfTerm();
	
	char *matrix = new char[nCols*nRows], *p = matrix;
	::memset(p, 0, nCols*nRows);
	for(i = F.begin(); i != F.end(); ++i) if(!i->isZero()) {
		Terms.getVector(*i, p);
		p += nCols;
	}
	
#define mat(r,c) matrix[(r)*nCols + (c)]
#define PUTMAT for(size_t rr = 0; rr < nRows; ++rr) {\
for(size_t cc = 0; cc < nCols; ++cc) std::cout << (mat(rr,cc)?'1':'0');\
std::cout << std::endl; } std::cout << std::endl;

//ProfNG("InMtx: " << nCols << "x" << nRows);
	// matrix operatrions
	size_t leadRow = 0;
	for(size_t iCol = 0; iCol < nCols; ++iCol) {
		// find pivot
		size_t iRow = leadRow;
		for(; iRow < nRows; ++iRow) if(mat(iRow, iCol)) break;
		
		// not found
		if(iRow >= nRows) continue;
		
		// found after leading row -> add the row at leading row
		if(iRow > leadRow) {
			for(size_t jCol = iCol; jCol < nCols; ++jCol) {
				mat(leadRow, jCol) ^= mat(iRow, jCol);
			}
		}
			
		// elimination
		for(iRow = 0; iRow < nRows; ++iRow) {
			if(iRow == leadRow) continue;
			if(mat(iRow, iCol)) {
				for(size_t jCol = iCol; jCol < nCols; ++jCol) {
					mat(iRow, jCol) ^= mat(leadRow, jCol);
				}
			}
		}
		if(++leadRow >= nRows) break;
	}
	// zero remove
	while(nRows > 0) {
		size_t iCol = 0;
		for(; iCol < nCols; ++ iCol) {
			if(matrix[(nRows -1)*nCols + iCol]) break;
		}
		if(iCol < nCols) break;
		--nRows;
	}
		
//ProfNG("OutMtx: " << nCols << "x" << nRows);
	p = matrix;
	for(size_t iRow = 0; iRow < nRows; ++iRow) {
		FT.push_back(F2poly(Terms, p));
		p += nCols;
	}
		
	delete [] matrix;
}

// ----------------------------------------------------------------------
//  Reduction Utilities
// ----------------------------------------------------------------------
bool
BackwardReduction(F2poly::Set const& F, F2poly::Set& G, PairList& P) {
	return true;
}

void
ReduceSimple(F2poly::Set& F) {
	bool changed;
	F2poly::Set::iterator i, j;
	do {
		changed = false;
		F.sort();
		for(i = F.begin(); i != F.end(); ) {
			if(i->isZero()) {
				F.erase(i++);
				continue;
			}
			for(j = F.begin(); j != F.end(); ++j) {
				if(j == i) continue;
				if(j->isZero()) continue;
				changed |= j->reducedBy(*i);
			}
			++i;
		}
	} while(changed);
}

struct CompNumTerms { bool operator()(F2poly const& l, F2poly const& r) {
	return l.getNumOfTerm() < r.getNumOfTerm();
}};

void
Resolve(F2poly::Set& F) {
	F2poly::Set G;
	F.sort(CompNumTerms());
	F2poly::Set::iterator iS = F.begin();
	for(; iS != F.end(); ++iS) {
		if(iS->getNumOfTerm() < 2) continue;
		if(iS->getNumOfTerm() == 2) {
			F2poly::terms k = iS->first(); ++k;
			if(!iS->LM().canBeDividedBy(*k)) continue;
			F2term x = iS->LM()/(*k);
			if(x.getDegree() < 2) continue;
			for(size_t i = 0; i < MAX_NUM_OF_VARIABLE; ++i) {
				if(x.hasVar(i)) {
					F2term y(*k); y.setVar(i);
					G.push_back(F2poly(*k));
					G.back() += y;
				}
			}
			continue;
		}
		for(size_t i = 0; i < MAX_NUM_OF_VARIABLE; ++i) {
			F2term const* px;
			int cn = 0;
			F2poly::terms k = iS->last(); --k;
			for(; k != iS->first(); --k) if(!k->hasVar(i)) {
				if(++cn > 1) break;
				px = &*k;
			}
			if(cn > 1) continue;
			if(!k->hasVar(i)) {
				if(++cn > 1) continue;
				px = &*k;
			}
			if(cn < 1) continue;
			F2term t(*px);
			G.push_back(F2poly(t));
			t.setVar(i); G.back() += t;
		}
	}
	F.splice(F.end(), G, G.begin(), G.end());
}

void
Unique(F2poly::Set& F, std::map<F2poly const*,F2poly const*>& pft) {
	F2poly::Set::iterator iS = F.begin(), i, j;
	for(; iS != F.end(); ++iS) if(!iS->isZero()) break;
	for(i = iS; i != F.end(); i = j) {
		for(j = i, ++j; j != F.end(); ++j) {
			if(*i != *j) break;
			j->zero();
			pft[&*j] = &*i;
		}
	}
	F.sort(CompNumTerms());
}


bool
ReduceNoHead1(F2poly::Set& F, bool uniq = false) {
	F.sort(CompNumTerms());
	F2poly::Set::iterator iS = F.begin();
	for(; iS != F.end(); ++iS) if(!iS->isZero()) break;
	
	bool changed = false;
	F2poly::Set::iterator i = iS, j;
	for(; i != F.end(); ++i) for(j = i, ++j; j != F.end(); ++j) {
		changed |= j->reducedNoHeadBy(*i);
	}
	F.sort(CompNumTerms());
	if(uniq) {
		for(j = F.end(), --j; j != iS; --j) {
			i = j; --i;
			if(*i == *j) j->zero();
		}
		F.sort(CompNumTerms());
	}
	return changed;
}

void
ReduceNoHead(F2poly::Set& F) {
	while(ReduceNoHead1(F));
}


// ----------------------------------------------------------------------
//  Reduction functions
// ----------------------------------------------------------------------

struct ReductionWithSimplify {
	int nEnter_, nReenter_, nMatch_;
	List FF, FE;
	
	// FF(whose back is incomplete F)  is one longer than FE!!
	F2poly
	MultSimplify(F2term const& t, F2poly const& f) {
		++nEnter_;
		// search (t', f') s.t.  t'|t, t' < t, and t'f' = tf,
		// 	and replace (t, f) with (t', f').
		//
		// inperfect (but fast) search for (t', f')s
		// 	t':the smaller, the better
		F2term lm = t*f.LM(), u;
		List::const_iterator i = FF.begin(), j = FE.begin();
		for(; j != FE.end(); ++j, ++i) {
			// j->beigin() has maximum degree
			if(!j->size()) continue;
			if(j->begin()->LM().getDegree()
				<= f.LM().getDegree()) continue;
			F2poly::Set::const_iterator ii = i->begin(), jj;
			for(; ii != i->end(); ++ii) {
				if(ii->LM() == f.LM()) continue;
				if(!ii->LM().canBeDividedBy(f.LM())) continue;
				if(!lm.canBeDividedBy(ii->LM())) continue;
				u = ii->LM()/f.LM();
				++nMatch_;
				F2poly uf = f; uf *= u;
				for(jj = j->begin(); jj != j->end(); ++jj) {
					if(uf.LM() == jj->LM()) break;
				}
				if(jj == j->end()) {
ProfNG("ouch\n");
					continue;
				}
				if(u == t) return *jj;
				++nReenter_;
				return MultSimplify(t/u, *jj);
			}
		}
		F2poly ret = f;
		return ret *= t;
	}

	void
	SymbolicPreprocess(PairList const& Pd, F2poly::Set const& G) {
Prof("SymbolicPre Enter");
		FF.push_back(F2poly::Set());
		F2poly::Set & F = FF.back();
	
		// F := {mult(Simplify(t, f, FF)) | (t, f) in Pd }
		for(PairList::const_iterator i = Pd.begin(); i!= Pd.end(); ++i){
			if(&i->f1_  == &i->f2_) { // f, S_i
				for(size_t j = 0; j < MAX_NUM_OF_VARIABLE; ++j){
					if(!i->f1_.LM().hasVar(j)) continue;
					F2term t1; t1.setVar(j);
					F.push_back(MultSimplify(t1, i->f1_));
				}
			} else { // f, g
				if(F2term::isPrimeTo(i->f1_.LM(), i->f2_.LM())){
					continue;
				}
				F2term lcm = F2term::getLCM(i->f1_.LM(),
							 i->f2_.LM());
				F.push_back(MultSimplify(lcm/i->f1_.LM(),
							 i->f1_));
				F.push_back(MultSimplify(lcm/i->f2_.LM(),
							 i->f2_));
			}
		}
		F.unique();
		//F.sort();
	
		// Done := HT(F),  Terms := T(F)
		F2poly::Set::iterator i;
		F2poly Done, Terms;
		for(i = F.begin(); i != F.end(); ++i) {
			Done.addUnique(i->LM());
			Terms.addUnique(*i);
		}
//ProfNG("\nHT(F) = " << Done << "\n");
	
		for(;;) {
			if(Done.getNumOfTerm() >= Terms.getNumOfTerm()) break;
			F2poly diff = Terms.notIn(Done);
			Done += diff;
			
			for(F2poly::terms k = diff.first(); k != diff.last(); ++k) {
				F2poly::Set::const_iterator j;
				for(j = G.begin(); j != G.end(); ++j) {
					if(k->canBeDividedBy(j->LM())) break;
				}
				if(j != G.end()) {
					F2term m = *k/j->LM();
//ProfNG("\nm = " << m << ", j = " << *j << "\n");
					F2poly h = MultSimplify(m, *j);
					F.push_back(h);
					Terms.addUnique(h);
				}
			}
		}
//ProfNG("\nTerms = " << Terms << "\n");
Prof("SymbolicPre Exit");
	}

	void
	operator()(PairList const& Pd, F2poly::Set const& G, F2poly::Set& FP) {
		// make F = FF.back()
		SymbolicPreprocess(Pd, G);
		F2poly::Set const& F = FF.back();
	
		// make FT = FE.back()
		FE.push_back(F2poly::Set());
		F2poly::Set& FT = FE.back();
		// @@ MinimumRowEchelonForm(F, FT);
	
		// make FP
		F2poly::Set::iterator i;
		F2poly::Set::const_iterator j;
		for(i = FT.begin(); i != FT.end(); ++i) {
			bool appendFP = true;
			for(j = F.begin(); j != F.end(); ++j) {
				if(i->LM() == j->LM()) {
					appendFP = false;
					break;
				}
			}
			// if(appendFP) FP.push_back(*i);
			// FE.back() :=  F~\F~+ is better!!
			if(appendFP) FP.splice(FP.end(), FT, i++), --i;
		}
	}
};


struct ReductionSimple {
	void
	SymbolicPreprocess(PairList const& Pd, F2poly::Set const& G,
		F2poly::Set& F, F2poly& Done) {
//Prof("SymbolicPre Enter");
	
		// F := {mult(Simplify(t, f, FF)) | (t, f) in Pd }
		for(PairList::const_iterator i = Pd.begin(); i!= Pd.end(); ++i){
			if(i->f1_.isZero()) continue;
			if(i->f2_.isZero()) continue;
			if(&i->f1_  == &i->f2_) { // f, S_i
				for(size_t j = 0; j < MAX_NUM_OF_VARIABLE; ++j){
					if(!i->f1_.LM().hasVar(j)) continue;
					F2term t1; t1.setVar(j);
					F.push_back(i->f1_);
					F.back().mulTopHigh(t1);
				}
				F.push_back(F2poly(i->f1_.LM()));
				const_cast<F2term&>(F.back().LM()).makeHigh();
				F.back() += i->f1_.LM();
			} else { // f, g
				if(F2term::isPrimeTo(i->f1_.LM(), i->f2_.LM())){
					continue;
				}
				F2term lcm = F2term::getLCM(i->f1_.LM(),
							 i->f2_.LM());
				F.push_back(i->f1_);
				F.back().mulTopHigh(lcm/i->f1_.LM());
				F.push_back(i->f2_);
				F.back().mulTopHigh(lcm/i->f2_.LM());
			}
		}
		F.unique();
		
		F2poly::Set::iterator iF;
		F2poly::Set::const_iterator iG;
		F2poly::terms k;
		F2poly Rem;
		for(iF = F.begin(); iF != F.end(); ++iF) {
			Done.addUnique(iF->LM());
			for(k = iF->first(), ++k; k != iF->last(); ++k) {
				Rem.addUnique(*k);
			}
		}
		while(Rem.getNumOfTerm() > 0) {
			int n = F.size();
			F2poly Top, Rep;
			for(k = Rem.first(); k != Rem.last(); ++k) {
				for(iG = G.begin(); iG != G.end(); ++iG) {
					if(iG->isZero()) continue;
					if(k->canBeDividedBy(iG->LM())) break;
				}
				if(iG == G.end()) {
					Done.addUnique(*k);
					continue;
				}
				F.push_back(*iG);
				Top.addUnique(*k);
				F.back().mulTopHigh(*k/iG->LM());
				Done.addUnique(F.back().LM());
			}
			for(iF = F.begin(); iF != F.end(); ++iF) {
				iF->makeHigh(Top);
			}
			F2poly Tmp;
			iF = F.begin(); while(--n >= 0) ++iF;
			for(; iF != F.end(); ++iF) {
				for(k = iF->first(); k != iF->last(); ++k) {
					if(!k->isHigh()) Tmp.addUnique(*k);
				}
			}
			Rem = Tmp;
		}
//Prof("SymbolicPre Exit");
	}
	
	void
	operator()(PairList const& Pd, F2poly::Set& G, F2poly::Set& FP) {
		F2poly::Set F;
		F2poly Terms;
		
		G.sort();
		SymbolicPreprocess(Pd, G, F, Terms);
		MinimumRowEchelonForm(F, Terms, FP);
	
		// make FP
		F2poly::Set::iterator i;
		for(i = FP.begin(); i != FP.end(); ) {
			if(i->LM().isHigh()) FP.erase(i++); else ++i;
		}
		ReduceSimple(FP);
	}
};


// ----------------------------------------------------------------------
//  Main Schemes of F4
// ----------------------------------------------------------------------
	
// assume G:null set
template<typename S_, typename U_, typename R_>
void AlgorithmF4(F2poly::Set const& F, 
	S_ Select, U_ Update, R_ Reduction, F2poly::Set& G) {
	PairList P;

	F2poly::Set::const_iterator i;
	for(i = F.begin(); i != F.end(); ++i) Update(G, P, *i);
	while(P.size() > 0) {
//ProfNG("\n loop start:\n");
//PUTG(G); P.PUT();
		PairList Pd;
		Select(P, Pd);
//ProfNG("\n selected :\n");
//Pd.PUT();
		F2poly::Set FP;
		Reduction(Pd, G, FP);
		for(i = FP.begin(); i != FP.end();++i) Update(G, P, *i);
ProfNG("\n #G = " << G.size() << ", #P = " << P.size() << "\n");
	}
//ProfNG("\n result \n");
//PUTG(G);
}



// ----------------------------------------------------------------------
//  instanciation
// ----------------------------------------------------------------------
void MakeGbaseBy(F2poly::Set& G) {
	F2poly::Set F;
	F.swap(G);
	Resolve(F);
	ReduceSimple(F);
	
	//ReductionWithSimplify R;
	ReductionSimple R;
	AlgorithmF4(F, 
		SelectMinDegLCM(),
		//SelectMinNTerms(),
		UpdateSimple(),
		//UpdateF5(),
		R, G);
//ProfNG("en = " << R.nEnter_ << 
//	", re = " << R.nReenter_ << ", ma = " << R.nMatch_);
	ReduceSimple(G);
ProfNG("\n");
}

#endif



} // end of namespace F4
} // end of namespace Hotaru

#endif // HOTARU_F4_H_
