#ifndef   HOTARU_F4_POLY_H_
#define   HOTARU_F4_POLY_H_
/**
	@file hotaru_f4_poly.h
	@brief Groebner Base(F4)
	@author $Author: dskw $
	@version $Revision: 1.16 $
	@date $Date: 2005/02/22 09:45:24 $
*/

#include <map>

// ======================================================================
//  container setting
// ======================================================================

#if defined(USE_MYLIST) || defined(USE_MYSET)
	#ifndef USE_MYLIST 
		#include "mylist.h"
	#endif
	#define POLF4_CONT mylist
	#define POLF4_OPT_ALLOC
#else
	#define POLF4_CONT std::list
	#ifdef USE_MYALLOC
		#define POLF4_OPT_ALLOC , myalloc<int> /**/
	#elif defiend(USE_BOOST)
		#define POLF4_OPT_ALLOC , boost::pool_allocator<int> /**/
	#else
		#define POLF4_OPT_ALLOC
	#endif
#endif

#ifdef MAIN_INSTANCE_HERE
mylistimp::mypool<int, POOL_NODE_NUM>
	mylistimp::cont_base<int, POOL_NODE_NUM>::pool_;
#endif


// ======================================================================
//  implementation
// ======================================================================

namespace Hotaru {
namespace F4 {

// ----------------------------------------------------------------------
//  debug settings
// ----------------------------------------------------------------------
#ifndef PROF
	#ifdef USE_PARAF4
		#define PROF 1
	#else
		#define PROF 2
	#endif
#endif
#if(PROF == 1)
	#define Prof(x)\
	std::cout << x << "\t\t#G = " << G.size() << " \t\r"; std::cout.flush()
	#define ProfNG(x)\
	std::cout << x << "\t\t\t\r"; std::cout.flush()
#elif(PROF == 2)
	#define Prof(x)\
	std::cout << x << "\t\t#G = " << G.size() << " \t" << std::endl
	#define ProfNG(x)\
	std::cout << x << "\t\t\t" << std::endl
#else
	#define Prof(x)
	#define ProfNG(x)
#endif


// ----------------------------------------------------------------------
//  main class
// ----------------------------------------------------------------------
#ifdef USE_HFE
extern int tdeg_;
#endif
struct PolySet {
	typedef std::map<F2term, int, std::greater<F2term> > Terms;
	typedef mylist<int> Poly;
	
	Terms terms_;
	std::vector<Terms::iterator> ix_;
	std::vector<Poly> polys_;
	
#ifndef USE_HL
	F2poly::Set F_;
#endif
	
	// --------------------------------------------------
	//  terms control
	// --------------------------------------------------
#ifdef USE_HL
	void
	unlowTerms() {
		for(size_t i = 0; i < ix_.size(); ++i) {
			if(ix_[i]->first.isLow()) {
				F2term t = ix_[i]->first;
				t.makeHigh();
				terms_.erase(ix_[i]);
				terms_[t] = i;
				ix_[i] = terms_.find(t);
			}
		}
#ifndef USE_SET
		remapAndSort();
#endif
	}
#endif
	
	void
	remapAndSort() {
		std::map<int, int> ip;
		Terms::iterator iT = terms_.begin();
		for(int i = 0; iT != terms_.end(); ++iT, ++i) {
			ip[iT->second] = i;
			iT->second = i;
			ix_[i] = iT;
		}
		for(size_t i = 0; i < polys_.size(); ++i) {
			Poly::iterator iK = polys_[i].begin();
			for(; iK != polys_[i].end(); ++iK) *iK = ip[*iK];
			polys_[i].sort(std::less<int>());
		}
	}
	
	// --------------------------------------------------
	//  poly access
	// --------------------------------------------------
	void
	addPoly(F2poly const& f) {
		if(f.isZero()) return;
#ifndef USE_HL
		F_.push_back(f);
#endif
		polys_.push_back(Poly());
		Poly &p = polys_.back();
		F2poly::terms k = f.first();
		for(; k != f.last(); ++k) {
			Terms::iterator iT = terms_.find(*k);
#ifdef USE_HL
			if(iT == terms_.end()) {
				F2term t(*k); t.makeHigh();
				iT = terms_.find(t);
			}
#endif
			if(iT == terms_.end()) {
				int n = ix_.size();
				terms_[*k] = n;
				ix_.push_back(terms_.find(*k));
				p.push_back(n);
			} else p.push_back(iT->second);
		}
	}
	
	size_t numPolys() const { return polys_.size(); }
	
	void
	getPoly(int i, F2poly& f) const {
		f.zero();
		Poly const& p = polys_[i];
		for(Poly::const_iterator iP = p.begin(); iP != p.end();
#ifndef USE_SET
			++iP) f.addTail(ix_[*iP]->first);
#else
			++iP) f += ix_[*iP]->first;
#endif
	}
	
	// --------------------------------------------------
	//  reduction
	// --------------------------------------------------
#ifdef USE_HL
	void
	makeHigh(int i) {
		Terms::iterator iT = ix_[i];
		F2term t = iT->first;
		t.makeHigh();
		terms_.erase(iT);
		terms_[t] = i;
		ix_[i] = terms_.find(t);
	}
	
	void
	makeLow(int i) {
		Terms::iterator iT = ix_[i];
		F2term t = iT->first;
		t.makeLow();
		terms_.erase(iT);
		terms_[t] = i;
		ix_[i] = terms_.find(t);
	}
#endif
	
	F2poly const*
	topReducibleBy(F2poly::Set const& G, int i) {
		F2term const& t = ix_[i]->first;
		for(F2poly::Set::const_reverse_iterator iG = G.rbegin();
			iG != G.rend(); ++iG) if(t.canBeDividedBy(iG->LM())) {
			return &*iG;
		}
		return 0;
	}
	
	F2poly const*
	topTrueReducibleBy(F2poly::Set const& G, int i) {
		F2term const& t = ix_[i]->first;
		for(F2poly::Set::const_reverse_iterator iG = G.rbegin();
			iG != G.rend(); ++iG) {
			if(t.canBeDividedBy(iG->LM()) && t != iG->LM()) {
				return &*iG;
			}
		}
		return 0;
	}
	
	static bool
	findTerm(Poly const& p, int i) {
		Poly::const_iterator iK = p.begin();
		for(; iK != p.end(); ++iK) {
			if(*iK > i) return false;
			if(*iK == i) return true;
		}
		return false;
	}
	
	static void
	polyPlus(Poly& pd, Poly const& ps) {
		Poly::iterator iD = pd.begin();
		Poly::const_iterator iS = ps.begin();
		while(iD != pd.end() && iS != ps.end()) {
			if(*iS < *iD) iD = pd.insert(iD, *iS++);
			else if(*iD < *iS) ++iD;
			else {
				pd.erase(iD++);
				++iS;
			}
		}
		if(iS != ps.end()) pd.insert(pd.end(), iS, ps.end());
	}

	size_t beginReduce_;
	
	bool
	reduceBy(F2poly::Set const& G, bool reset = true) {
		bool changed = false;
		
		if(reset) beginReduce_ = 0;
		
		// prepare reducers form G
		F2poly const* pf;
		for(size_t i = beginReduce_; i < ix_.size(); ++i) {
#ifdef USE_HL
			if(!(pf = topReducibleBy(G, i))) continue;
			F2term t = ix_[i]->first/pf->LM();
			makeHigh(i);
			F2poly f(*pf);
			addPoly(f.mulTopHigh(t));
#else
			if(!(pf = topReducibleBy(G, i)) &&
			   !(pf = topTrueReducibleBy(F_, i))) continue;
			F2poly f(*pf); f *= ix_[i]->first/pf->LM();
			addPoly(f);
#endif
			changed = true;
		}
		if(!changed) {
			for(size_t i0 = 1; i0 < F_.size(); ++i0) {
				int it = *polys_[i0].begin();
				for(size_t i1 = 0; i1 < i0; ++i1) {
					if(*polys_[i1].begin() == it) {
						changed = true;
						break;
					}
				}
				if(changed) break;
			}
		}
		if(!changed) return changed;
		
		remapAndSort();
		
#ifdef USE_HL
		// calculate minimum index of non-high term
		size_t minLo = 0;
		while(minLo < ix_.size() && ix_[minLo]->first.isHigh()) ++minLo;
#else
		// calculate minimum index of non-G-top-reducible term
		size_t minLo = 0;
		while(minLo < ix_.size() && (topReducibleBy(G, minLo) || 
			topTrueReducibleBy(F_, minLo))) ++minLo;
		
#endif
		
#if 0
std::cout << "Input = [" << std::endl;
for(size_t ii = 0; ii < polys_.size(); ++ii) {
F2poly f; getPoly(ii, f);
std::cout << "  G[" << ii << "] = " << f << std::endl;
}
std::cout << "  Terms = [ ";
for(size_t ii = 0; ii < ix_.size(); ++ii) {
std::cout << ix_[ii]->first << " ";
}
std::cout << "]\n];" << std::endl;
#endif
		
		// row echelon form
#if 0
		size_t leadRow = 0, nCols = ix_.size(), nRows = polys_.size();
		for(size_t iCol = 0; iCol < nCols; ++iCol) {
			// search for a pivot
			size_t pvRow = leadRow;
			for(; pvRow < nRows; ++pvRow) {
				if(findTerm(polys_[pvRow], iCol)) break;
			}
			if(pvRow >= nRows) continue;
			if(pvRow > leadRow) {
				polys_[pvRow].swap(polys_[leadRow]);
			}
			
			// elimination
			for(size_t iRow = 0; iRow < nRows; ++iRow) {
				if(iRow == leadRow) continue;
				if(!findTerm(polys_[iRow], iCol)) continue;
				polyPlus(polys_[iRow], polys_[leadRow]);
			}
			if(++leadRow >= nRows) break;
		}
#else
		// set matrix
		size_t leadRow = 0, nCols = ix_.size(), nRows = polys_.size();
//#define BTYPE unsigned char
#define BTYPE unsigned int
		size_t const bits = 8*sizeof(BTYPE);
		size_t nColBlks = (nCols + bits - 1)/bits;
#define mat(x, y) matrix[(x) + (y)*nColBlks]
#define cmat(x, y) ((unsigned char*)(matrix + (x)/bits + (y)*nColBlks))[(x)/8]
std::cerr << "matrix: " << sizeof(BTYPE) << " of "
	<< nColBlks << " x " << nRows << std::endl;
		BTYPE *matrix = new BTYPE[nColBlks*nRows];
		::memset(matrix, 0, nColBlks*nRows*sizeof(BTYPE));
		for(size_t iRow = 0; iRow < nRows; ++iRow) {
			Poly& p = polys_[iRow];
			for(Poly::iterator iK = p.begin(); iK != p.end(); ++iK){
				mat(*iK/bits, iRow) |= (BTYPE)1 << *iK%bits;
			}
		}
		polys_.resize(0);
		
		// solve matrix
		for(size_t iCol = 0; iCol < nCols; ++iCol) {
#if 1
			if(iCol&7 == 0 && iCol < nCols - 128) {
				for(size_t iR = leadRow; iR < nRows; ++iR){
					unsigned char ch = cmat(iCol, iR);
					if(!ch) continue;
					for(size_t jR = 0; jR < nRows; ++jR) {
						if(jR == iR) continue;
						if(cmat(iCol, jR) != ch) continue;
						for(size_t jB = iCol/bits;
							jB < nColBlks; ++jB) {
							mat(jB, jR) ^= mat(jB, iR);
						}
					}
				}
			}
#endif
			// search for a pivot
			size_t pvRow = leadRow;
			for(; pvRow < nRows; ++pvRow) {
				if(mat(iCol/bits, pvRow)&((BTYPE)1<<iCol%bits)) break;
			}
			if(pvRow >= nRows) continue;
			if(pvRow > leadRow) {
				// add pvRow to leadRow
				for(size_t jB = iCol/bits; jB < nColBlks; ++jB){
					std::swap(mat(jB, leadRow), mat(jB, pvRow));
				}
			}
			
			// elimination
			for(size_t iRow = 0; iRow < nRows; ++iRow) {
				if(iRow == leadRow) continue;
				if(!(mat(iCol/bits, iRow)&((BTYPE)1<<iCol%bits))) continue;
				for(size_t jB = iCol/bits; jB < nColBlks; ++jB){
					mat(jB, iRow) ^= mat(jB, leadRow);
				}
			}
			if(++leadRow >= nRows) break;
		}
		
		// get matrix
		size_t iMinB = 0, iMinLB = minLo/bits;
		for(size_t iRow = 0; iRow < nRows; ++iRow) {
			bool isHigh = false;
			for(size_t iB = iMinB; iB < nColBlks; ++iB) {
				if(!mat(iB, iRow)) { ++iMinB; continue; }
				if(iB < iMinLB) isHigh = true;
				break;
			}
			if(isHigh) continue;
			if(iMinB >= nColBlks) break;
#ifndef USE_HL
			BTYPE t = 1, c = mat(iMinB, iRow);
			for(size_t jB = 0; jB < bits; ++jB, t += t){
				if(!(c & t)) continue;
				if(topReducibleBy(G, iMinB*bits+jB) ||
				   topTrueReducibleBy(F_, iMinB*bits+jB)) {
					isHigh = true;
				}
				break;
			}
			if(isHigh) continue;
#endif
			
			polys_.push_back(Poly());
			Poly &p = polys_.back();
			for(size_t iB = iMinB; iB < nColBlks; ++iB) {
				BTYPE ch = mat(iB, iRow);
				if(!ch) continue;
				BTYPE tst = 1;
				for(size_t jB = 0; jB < bits; ++jB, tst += tst){
					if(!(ch & tst)) continue;
					if(iB*bits + jB >= nCols) break;
					p.push_back(iB*bits + jB);
				}
			}
		}
		delete [] matrix;
#undef cmat
#undef mat
#undef BTYPE
#endif

#if 0
std::cout << "Output = [" << std::endl;
for(size_t ii = 0; ii < polys_.size(); ++ii) {
F2poly f; getPoly(ii, f);
std::cout << "  G[" << ii << "] = " << f << std::endl;
}
std::cout << "];" << std::endl;
#endif
		
		// skip polys which have high term(s) or zero
		nRows = polys_.size();
		size_t nNewRows = 0;
		for(size_t iRow = 0; iRow < nRows; ++iRow) {
			Poly& p = polys_[iRow];
			if(!p.size()) continue;
			if(*p.begin() < (int)minLo) continue;
			if(nNewRows != iRow) std::swap(p, polys_[nNewRows]);
			++nNewRows;
		}
		polys_.resize(nNewRows);
		
//std::cout << "Reduced = [" << std::endl;
//for(size_t ii = 0; ii < polys_.size(); ++ii) {
//F2poly f; getPoly(ii, f);
//std::cout << "  G[" << ii << "] = " << f << std::endl;
//}
//std::cout << "];" << std::endl;
		return changed;
	}
	
	// --------------------------------------------------
	//  set access
	// --------------------------------------------------
	void
	addSet(F2poly::Set const& G) {
		F2poly::Set::const_iterator iG = G.begin();
		for(; iG != G.end(); ++iG) addPoly(*iG);
		beginReduce_ = 0;
	}
	
	void
	getSet(F2poly::Set& G) {
		for(size_t i = 0; i < polys_.size(); ++i) {
			G.push_back(F2poly());
			getPoly(i, G.back());
			//if(G.back().LM().getDegree() > tdeg_) G.pop_back();
		}
	}
	
#ifdef USE_HL
	void
	addSetSelf(F2poly::Set const& G) {
		F2poly::Set::const_iterator iG = G.begin();
		for(; iG != G.end(); ++iG) {
			if(terms_.find(iG->LM()) == terms_.end()) {
				int n = ix_.size();
				terms_[iG->LM()] = n;
				ix_.push_back(terms_.find(iG->LM()));
			}
		}
		beginReduce_ = ix_.size();
		for(iG = G.begin(); iG != G.end(); ++iG) addPoly(*iG);
	}
#endif
};


// ----------------------------------------------------------------------
//  poly reduction
// ----------------------------------------------------------------------
struct lessLMnumTerms {
	bool operator()(F2poly const& l, F2poly const& r) const {
		int cmp = l.LM().isGreater(r.LM());
		if(cmp) return cmp < 0;
		return l.getNumOfTerm() < r.getNumOfTerm();
	}
};

void
MinimizeS(F2poly::Set& G) {
	for(F2poly::Set::iterator iG = G.begin(); iG != G.end();) {
		if(iG->isZero()) G.erase(iG++);
		else ++iG;
	}
	G.sort(lessLMnumTerms());
	F2poly::Set::iterator iG = G.begin(), jG;
	for(; iG != G.end(); ++iG) {
		for(jG = iG, ++jG; jG != G.end(); ++jG) {
			if(jG->LM().canBeDividedBy(iG->LM())) {
				F2poly f(*iG);
				f *= jG->LM()/f.LM();
				*jG -= f;
			}
		}
	}
}

void
Minimize(F2poly::Set& G) {
	bool changed = true;
	while(changed) {
		for(F2poly::Set::iterator iG = G.begin(); iG != G.end();) {
			if(iG->isZero()) G.erase(iG++);
			else ++iG;
		}
		G.sort(lessLMnumTerms());
		changed = false;
		F2poly::Set::iterator iG = G.begin(), jG;
		for(; iG != G.end(); ++iG) {
			for(jG = iG, ++jG; jG != G.end(); ++jG) {
				if(jG->LM().canBeDividedBy(iG->LM())) {
					F2poly f(*iG);
					f *= jG->LM()/f.LM();
					*jG -= f;
					changed = true;
				}
			}
		}
	}
}

void
MinimizeGbase(F2poly::Set& G) {
	G.sort(lessLMnumTerms());
	for(F2poly::Set::iterator iG = G.begin(); iG != G.end();) {
		if(iG->isZero()) G.erase(iG++);
		else ++iG;
	}
	F2poly::Set::iterator iG = G.begin(), jG;
	for(; iG != G.end(); ++iG) {
		for(jG = iG, ++jG; jG != G.end();) {
			if(jG->LM().canBeDividedBy(iG->LM())) G.erase(jG++);
			else ++jG;
		}
	}
#if 0
std::cout << "Minimal = [" << std::endl;
int ii = 0;
for(F2poly::Set::iterator iG = G.begin(); iG != G.end(); ++iG) {
std::cout << "  G[" << ii++ << "] = " << *iG << std::endl;
}
std::cout << "];" << std::endl;
#endif
}

#ifdef USE_HL
void
ReduceMinimalSet(F2poly::Set& F) {
	PolySet G;
	G.addSetSelf(F);
	G.reduceBy(F, false);
	G.unlowTerms();
	F.clear();
	G.getSet(F);
}

void
ReduceS(F2poly::Set& G) {
	MinimizeS(G);
	ReduceMinimalSet(G);
}

void
Reduce(F2poly::Set& G) {
	Minimize(G);
	ReduceMinimalSet(G);
}
#else
void Reduce(F2poly::Set& F) {
	for(;;) {
		PolySet G;
		G.addSet(F);
		if(!G.reduceBy(F2poly::Set())) break;
		F.clear();
		G.getSet(F);
	}
}
void ReduceS(F2poly::Set& F) {
	PolySet G;
	G.addSet(F);
	G.reduceBy(F2poly::Set());
	F.clear();
	G.getSet(F);
}
void ReduceMinimalSet(F2poly::Set& F) { ReduceS(F);}
#endif

void
ReduceGbase(F2poly::Set& G) {
	MinimizeGbase(G);
	ReduceMinimalSet(G);
}


} // end of namespace F4
} // end of namespace Hotaru

#endif // HOTARU_F4_POLY_H_
