#ifndef MIE_THREAD_H_
#define MIE_THREAD_H_
/**
	@file mie_thead.h
	@brief åɤδʰץ饹
	@author $Author: dskw $
	@version $Revision: 1.2 $
	@date $Date: 2005/02/15 17:46:33 $
	@note LinuxǤǤCriticalSection == mutex
*/

#ifdef _WIN32
	#include <windows.h>
	#include <process.h>
#elif defined(__GNUC__)
	#include <unistd.h>
	#include <pthread.h>
#else
	#error "not support"
#endif

namespace MIE {

/*
	Ƶmutex°
	̾ϥ֥Ȥååɤ٥å硤֥å
	뤬Υե饰trueˤȡ󥿤䤷ƼȤϥ롼롥
*/
const bool isRecursiveMutex = false;

#ifdef _WIN32

	#ifndef _MT
		#error "Compile under multithread mode runtime"
	#endif

	typedef HANDLE mie_thread_t;
	#define MIE_INFINITE INFINITE
	typedef void THREAD_ENTRY_TYPE(void*);

	/**
		åɥϥɥ
		@param hdl [out] åɥϥɥ
		@param entryFct [in] åɳϥȥ
		@param arg [in] 桼ѥ᡼
		@retval 0 ｪλ
	*/
	static inline int THREAD_Begin(mie_thread_t& hdl, THREAD_ENTRY_TYPE entryFct, void *arg)
	{
		hdl = (mie_thread_t)_beginthread(entryFct, 0, arg);
		return hdl == (HANDLE)-1;
	}

	/**
		åɽλԤ
		@param hdl [in] åɥϥɥ
		@param timeoutMsec [in] λԤ(msec)
	*/
	static inline int THREAD_Join(mie_thread_t hdl, int timeoutMsec)
	{
		return WaitForSingleObject(hdl, timeoutMsec) != WAIT_OBJECT_0;
	}

	/**
		Ԥ
		@param [in] msec Ԥ(msec)
	*/
	static inline void SleepMsec(int msec)
	{
		Sleep(msec);
	}

	typedef CRITICAL_SECTION mie_cs_t;
	static inline void CS_Init(mie_cs_t& cs) { InitializeCriticalSection(&cs); }
	static inline void CS_Lock(mie_cs_t& cs) { EnterCriticalSection(&cs); }
	static inline void CS_Unlock(mie_cs_t& cs) { LeaveCriticalSection(&cs); }
	static inline void CS_Term(mie_cs_t& cs) { DeleteCriticalSection(&cs); }

	#define mie_mutex_t HANDLE
	static inline void MUTEX_Init(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			mutex = CreateMutex(NULL, FALSE, NULL);
		} else {
			mutex = CreateSemaphore(NULL /* no security */, 1 /*  */, 0x7FFFFFFF /*  */, NULL /* ̾ʤ */);
		}
	}
	static inline void MUTEX_Lock(mie_mutex_t& mutex) { WaitForSingleObject(mutex, INFINITE); }
	static inline void MUTEX_Unlock(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			ReleaseMutex(mutex);
		} else {
			ReleaseSemaphore(mutex, 1, NULL);
		}
	}
	static inline void MUTEX_Term(mie_mutex_t& mutex) { CloseHandle(mutex); }

#elif defined(__GNUC__)

	typedef pthread_t mie_thread_t;
	typedef void* THREAD_ENTRY_TYPE(void*);

	/**
		åɥϥɥ
		@param hdl [out] åɥϥɥ
		@param entryFct [in] åɳϥȥ
		@param arg [in] 桼ѥ᡼
		@retval 0 ｪλ
	*/
	static inline int THREAD_Begin(mie_thread_t& hdl, THREAD_ENTRY_TYPE entryFct, void *arg)
	{
		return pthread_create(&hdl, NULL, entryFct, arg) != 0;
	}

	/**
		åɽλԤ
		@param hdl [in] åɥϥɥ
		@param timeoutMsec [in] λԤ(msec)(LinuxǤϤΥѥ᡼̵)
	*/
	static inline int THREAD_Join(mie_thread_t hdl, int timeoutMsec)
	{
		return pthread_join(hdl, NULL);
	}

	/**
		Ԥ
		@param [in] msec Ԥ(msec)
	*/
	static inline void SleepMsec(int msec)
	{
		usleep(msec * 1000);
	}

	typedef pthread_mutex_t mie_cs_t;
	static inline void CS_Init(mie_cs_t& cs) { pthread_mutex_init(&cs, NULL); }
	static inline void CS_Lock(mie_cs_t& cs) { pthread_mutex_lock(&cs); }
	static inline void CS_Unlock(mie_cs_t& cs) { pthread_mutex_unlock(&cs); }
	static inline void CS_Term(mie_cs_t& cs) { pthread_mutex_destroy(&cs); }

	typedef pthread_mutex_t mie_mutex_t;
	static inline void MUTEX_Init(mie_mutex_t& mutex)
	{
#ifdef __linux__
		if (isRecursiveMutex) {
			pthread_mutexattr_t mutexattr;
			pthread_mutexattr_init(&mutexattr);
			/* pthreadɸ */
			pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE_NP);
			pthread_mutex_init(&mutex, &mutexattr);
		} else {
#endif
			pthread_mutex_init(&mutex, NULL);
#ifdef __linux__
		}
#endif
	}
	static inline void MUTEX_Lock(mie_mutex_t& mutex) { pthread_mutex_lock(&mutex); }
	static inline void MUTEX_Unlock(mie_mutex_t& mutex) { pthread_mutex_unlock(&mutex); }
	static inline void MUTEX_Term(mie_mutex_t& mutex) { pthread_mutex_destroy(&mutex); }

#else
	#error "not supported"
#endif

/**
	å饹
*/
class ThreadBase {
protected:
	mie_thread_t threadHdl_;

	/*
		åɥ롼ץ󥿥ե
	*/
	static void threadLoopIF(void *arg)
	{
		ThreadBase* main = reinterpret_cast<ThreadBase*>(arg);
		main->threadEntry();
	}
public:
	/**
		åɤͥ
	*/
	enum Priority {
		HIGHEST, /* Ǥ⤤ͥ */
		HIGH, /* ⤤ͥ */
		NORMAL, /* ̾ */
		LOW /* 㤤 */
	};
	virtual void threadEntry() = 0; //!< åɳϥȥݥ
	virtual void stopThread() = 0; //!< åɤλ

	/**
		󥹥ȥ饯
	*/
	ThreadBase()
		: threadHdl_(0)
	{
	}

	/**
		åɳ
	*/
	bool beginThread()
	{
		return THREAD_Begin(threadHdl_, (THREAD_ENTRY_TYPE*)threadLoopIF, this) == 0;
	}

	/**
		åɤνλԤ
		@param waitMsec [in] ॢȻ(ǥե2)
	*/
	bool joinThread(int waitMsec = 2000)
	{
		if (threadHdl_) {
			if (THREAD_Join(threadHdl_, waitMsec)) {
				return false;
			}
		}
		return true;
	}
	/**
		åɤ̤ͥѹ
		@param priority [in] ͥ
	*/
	bool setThreadPriority(Priority priority)
	{
#ifdef WIN32
		int priorityValue;
		HANDLE hdl = GetCurrentThread();
		switch (priority) {
		case HIGHEST:
			priorityValue = THREAD_PRIORITY_HIGHEST;
		case HIGH:
			priorityValue = THREAD_PRIORITY_ABOVE_NORMAL;
			break;
		case NORMAL:
			priorityValue = THREAD_PRIORITY_NORMAL;
			break;
		case LOW:
			priorityValue = THREAD_PRIORITY_BELOW_NORMAL;
			break;
		default:
			return false;
		}
		SetThreadPriority(hdl, priorityValue);
#endif
		return true;
	}
	/**
		ץ̤ͥѹ
	*/
	static inline bool setProcessPriority(Priority priority)
	{
#ifdef WIN32
		DWORD priorityClass;
		HANDLE hdl = GetCurrentProcess();
		switch (priority) {
		case HIGHEST:
			priorityClass = HIGH_PRIORITY_CLASS;
		case HIGH:
			priorityClass = ABOVE_NORMAL_PRIORITY_CLASS;
			break;
		case NORMAL:
			priorityClass = NORMAL_PRIORITY_CLASS;
			break;
		case LOW:
			priorityClass = BELOW_NORMAL_PRIORITY_CLASS;
			break;
		default:
			return false;
		}
		SetPriorityClass(hdl, priorityClass);
		return true;
#else
		return true;
#endif
	}
};

/**
	ڤ¾楯饹
*/
class CriticalSection {
public:
	CriticalSection()
	{
		CS_Init(cs_);
	}
	~CriticalSection()
	{
		CS_Term(cs_);
	}
	inline void lock()
	{
		CS_Lock(cs_);
	}
	inline void unlock()
	{
		CS_Unlock(cs_);
	}
private:
public:
	mie_cs_t cs_;
};

#ifdef _MSC_VER

/**
	Ūʱ黻
*/
inline void AtomicAdd(CriticalSection& cs, int* x, int y)
{
	InterlockedExchangeAdd((long*)x, y);
}

inline void AtomicInc(CriticalSection& cs, int* x)
{
	InterlockedIncrement((long*)x);
}

#else

/**
	Ūʱ黻
*/
inline void AtomicAdd(CriticalSection& cs, int* x, int y)
{
	cs.lock();
	*x += y;
	cs.unlock();
}

inline void AtomicInc(CriticalSection& cs, int* x)
{
	AtomicAdd(cs, x, 1);
}

#endif

/**
	ưå
*/
class AutoLock {
public:
	AutoLock(CriticalSection &cs)
		: cs_(cs)
	{
		cs_.lock();
	}
	~AutoLock()
	{
		cs_.unlock();
	}
private:
	CriticalSection& cs_;
};

} // end of namespace

#endif /* MIE_THREAD_H_ */
