#ifndef   MY_ALLOC_H_
#define   MY_ALLOC_H_
/**
	@file myalloc.h
	@brief alloc with pool
	@author $Author: dskw $
	@version $Revision: 1.5 $
	@date $Date: 2005/02/18 07:28:43 $
 */

#include <iostream>
#include <new>
#include <vector>

namespace myallocimp {

template<typename T_, int num_, size_t size_ = sizeof(T_)>
struct mypool {
	typedef union tagE { char mem_[size_]; union tagE* ptr_; } Elem_;
	std::vector<Elem_*> vpool_;
	Elem_ *first_;
	mypool() : first_(0) {
		vpool_.reserve(100);
	}
	~mypool() {
		for(size_t i = 0; i < vpool_.size(); ++i) delete vpool_[i];
	}
	Elem_*
	alloc_new_pool() {
std::cerr << "num pools[" << num_ << ", "
	<< sizeof(Elem_) << "]: " << vpool_.size()+1 << std::endl;
		Elem_ *tmp = new Elem_[num_];
		vpool_.push_back(tmp);
		for(int i = 1; i < num_; ++i) tmp[i-1].ptr_ = &tmp[i];
		tmp[num_ - 1].ptr_ = 0;
		return tmp;
	}
	
	void* alloc() {
		if(!first_) first_ = alloc_new_pool();
		Elem_* ret = first_;
		first_ = first_->ptr_;
		return ret;
	}
	void dealloc(void *p) {
		Elem_ *pp = (Elem_*)p;
		pp->ptr_ = first_;
		first_ = pp;
	}
};

}

template<typename T_, int num_ = Hotaru::POOL_NODE_NUM>
struct myalloc {
	typedef size_t    size_type;
	typedef ptrdiff_t difference_type;
	typedef T_*        pointer;
	typedef const T_*  const_pointer;
	typedef T_&        reference;
	typedef const T_&  const_reference;
	typedef T_         value_type;
	template <typename U_>
		struct rebind { typedef myalloc<U_> other; };
	pointer address(reference x) const { return &x; }
	const_pointer address(const_reference x) const { return &x; }
	static pointer allocate(size_type n) {
		assert(n == 1); // mypool's spec
		return (pointer)pool_.alloc();
	}
	static void deallocate(pointer p, size_type n) {
		assert(n == 1); // mypool's spec
		pool_.dealloc(p);
	}
	size_type max_size() const { return (size_t)-1 / sizeof(T_); }
	void construct(pointer p, const T_& val) { new ((void*)p) T_(val); }
	void destroy(pointer p) { p->~T_(); }
	
	static myallocimp::mypool<T_, num_> pool_;
};

#ifndef NOUSE_MYALLOC_TRAITS
namespace std {

template<typename T_, typename U_>
struct _Alloc_traits<T_, myalloc<U_> >
{
	static const bool _S_instanceless = true;
	typedef myalloc<T_> allocator_type;
	typedef myalloc<T_> _Alloc_type;
};

}
#endif

  
#endif // MY_ALLOC_H_
