#ifndef   MY_LIST_H_
#define   MY_LIST_H_
/**
	@file mylist.h
	@brief list with pool
	@author $Author: dskw $
	@version $Revision: 1.7 $
	@date $Date: 2005/02/18 07:28:43 $
 */

#include <iostream>
#include<new>

using namespace Hotaru;

namespace mylistimp {

struct node_base { node_base *next_, *prev_; };
template<typename T_>
struct node : public node_base { T_ data_; };

struct iter_base {
	node_base *node_;
	iter_base(){}
	iter_base(node_base *p) : node_(p) {}
	void forward() { node_ = node_->next_; }
	void backward() { node_ = node_->prev_; }
	bool operator==(iter_base const& i) const { return node_ == i.node_; }
	bool operator!=(iter_base const& i) const { return node_ != i.node_; }
};

template<typename T_, typename Ref_, typename Ptr_>
struct iter : public iter_base {
	typedef iter<T_, Ref_, Ptr_> Self_;
	typedef node<T_> Node_;
	
	iter(Node_* p) : iter_base(p) {}
	iter() {}
	iter(iter const& i) : iter_base(i.node_) {}
	
	Ref_ operator*() const { return static_cast<Node_*>(node_)->data_; }
	Ptr_ operator->() const { return &(operator*()); }
	Self_& operator++() { forward(); return *this; }
	Self_& operator--() { backward(); return *this; }
	Self_ operator++(int) { Self_ tmp = *this; forward(); return tmp; }
	Self_ operator--(int) { Self_ tmp = *this; backward(); return tmp; }
};

template<typename T_, int num_, size_t size_ = sizeof(node<T_>)>
struct mypool {
	typedef union tagE { char mem_[size_]; union tagE* ptr_; } Elem_;
	std::vector<Elem_*> vpool_;
	Elem_ *first_;
	mypool() : first_(0) {
		vpool_.reserve(100);
	}
	~mypool() {
		for(size_t i = 0; i < vpool_.size(); ++i) delete vpool_[i];
	}
	Elem_*
	alloc_new_pool() {
std::cerr << "num pools[" << num_ << ", "
	<< sizeof(Elem_) << "]: " << vpool_.size()+1 << std::endl;
		Elem_ *tmp = new Elem_[num_];
		vpool_.push_back(tmp);
		for(int i = 1; i < num_; ++i) tmp[i-1].ptr_ = &tmp[i];
		tmp[num_ - 1].ptr_ = 0;
		return tmp;
	}
	
	void* alloc() {
		if(!first_) first_ = alloc_new_pool();
		Elem_* ret = first_;
		first_ = first_->ptr_;
		return ret;
	}
	void dealloc(void *p) {
		Elem_ *pp = (Elem_*)p;
		pp->ptr_ = first_;
		first_ = pp;
	}
};


template<typename T_, int num_, size_t size_ = sizeof(node<T_>)>
struct cont_base {
	typedef node<T_> Node_;
	Node_ *node_;
	
	cont_base() : node_(get_node())
		 { node_->next_ = node_; node_->prev_ = node_; }
	~cont_base() { clear(); put_node(node_); } 
	void clear() {
		Node_* cur = static_cast<Node_*>(node_->next_);
		while(cur != node_) {
			Node_ *tmp = cur;
			cur = static_cast<Node_*>(cur->next_);
			tmp->data_.~T_();
			put_node(tmp);
		}
		node_->next_ = node_;
		node_->prev_ = node_;
        }

	static mypool<T_, num_, size_> pool_;
	
	static Node_* get_node() { return (Node_*)pool_.alloc(); }
	static void put_node(Node_* p) { pool_.dealloc(p); }
};

}

template<typename T_, int num_ = POOL_NODE_NUM>
struct mylist : public mylistimp::cont_base<T_, num_> {
	typedef mylistimp::cont_base<T_, num_> Base_;
	typedef mylistimp::iter<T_, T_ const&, T_ const*> const_iterator;
	typedef mylistimp::iter<T_, T_&, T_*> iterator;
	typedef mylistimp::node<T_> Node_;
	using Base_::node_;
	
	// ctor/dtor -- with no allocator
	// unsupport : 
	//   explicit ctor(size_t n, T_ const& v = T_())
	//   ctor(InIter f, InIter l)
	mylist() {}
	mylist(mylist const& r) { insert(begin(), r.begin(), r.end()); }
	~mylist() {}
	
	// op=/assign
	// unsupport :
	//   assign(*)
	mylist& operator=(mylist const& r) {
		if(this != &r) {
			iterator f1 = begin();
			iterator l1 = end();
			const_iterator f2 = r.begin();
			const_iterator l2 = r.end();
			while(f1 != l1 && f2 != l2) *f1++ = *f2++;
			if(f2 == l2) erase(f1, l1);
			else insert(l1, f2, l2);
		}
		return *this;
	}
	
	// get_allocator -- unsupport
	
	// begin/end -- reverse:unsupport
        const_iterator begin() const
		{ return static_cast<Node_*>(node_->next_); }
        iterator begin() { return static_cast<Node_*>(node_->next_); }
        const_iterator end() const { return static_cast<Node_*>(node_); }
        iterator end() { return static_cast<Node_*>(node_); }
	
	// empty/size -- resize/max_size:unsupport
	bool empty() const { return node_->next_ == node_; }
	size_t size() const {
		size_t n = 0;
		const_iterator first = begin();
		while(first != end()) { ++first; ++n; }
		return n;
 	}
	
	// front/back/push/pop
	// -- only push_back(T_ const&) support
	// -- supporting others is easy
	void push_back(T_ const& r) { insert(end(), r); }
	
	// insert
	iterator insert(iterator i, T_ const& r) {
		Node_ *tmp = get_node();
		new (static_cast<void*>(&tmp->data_)) T_(r);
		tmp->next_ = i.node_;
		tmp->prev_ = i.node_->prev_;
		i.node_->prev_->next_ = tmp;
		i.node_->prev_ = tmp;
		return tmp;
	}
	// const_iterator -> template<typename InIter_> will work
	void insert(iterator i, const_iterator f, const_iterator l) {
		for(; f != l; ++f) insert(i, *f);
	}
	
	// erase
	iterator erase(iterator i) {
		mylistimp::node_base *next = i.node_->next_;
		mylistimp::node_base *prev = i.node_->prev_;
		Node_* node = static_cast<Node_*>(i.node_);
		prev->next_ = next;
		next->prev_ = prev;
		node->data_.~T_();
		put_node(node);
		return static_cast<Node_*>(next);
	}
	iterator erase(iterator f, iterator l) {
		while(f != l) erase(f++);
		return l;
	}
	
	// swap
	void swap(mylist& r) { std::swap(node_, r.node_); }
	
	// clear -- see baseclass
	
	// splice
	//   unsupported:
	//     void splice(iterator, mylist&)
	//     void splice(iterator, mylist&, iterator, iterator)
	void splice(iterator p, mylist&, iterator i) {
		if(p.node_ == i.node_) return;
		if(p.node_ == i.node_->next_) return;
		
		p.node_->prev_->next_ = i.node_;
		i.node_->prev_->next_ = i.node_->next_;
		i.node_->next_->prev_ = i.node_->prev_;
		i.node_->next_ = p.node_;
		i.node_->prev_ = p.node_->prev_;
		p.node_->prev_ = i.node_;
	}
	
	// remove -- unsupport
	
	// merge/sort/unique
	//   unsupported:
	//     special(<) version of merge/sort
	//     unique
	template<typename Compare_>
	void sort(Compare_ const& comp) {
		if(node_->next_ == node_) return; //empty
		if(node_->next_->next_ == node_) return; //one-node
		mylist carry, counter[64];
		int fill = 0;
		while(!empty()) {
			carry.splice(carry.begin(), *this, begin());
			int i = 0;
			while(i < fill && !counter[i].empty()) {
				counter[i].merge(carry, comp);
				carry.swap(counter[i++]);
			}
			carry.swap(counter[i]);
			if(i == fill) ++fill;
		}
		for(int i = 1; i < fill; ++i) {
			counter[i].merge(counter[i-1], comp);
		}
		swap(counter[fill-1]);
	}
	template<typename Compare_>
	void merge(mylist& r, Compare_ comp) {
		iterator f1 = begin();
		iterator l1 = end();
		iterator f2 = r.begin();
		iterator l2 = r.end();
		
		while(f1 != l1 && f2 != l2) {
			if(comp(*f2, *f1)) {
				iterator next = f2; ++next;
				splice(f1, *this, f2);
				f2 = next;
			} else ++f1;
		}
		if(f2 != l2) {
			l2.node_->prev_->next_ = l1.node_;
			f2.node_->prev_->next_ = l2.node_;
			l1.node_->prev_->next_ = f2.node_;
			mylistimp::node_base *tmp = l1.node_->prev_;
			l1.node_->prev_ = l2.node_->prev_;
			l2.node_->prev_ = f2.node_->prev_;
			f2.node_->prev_ = tmp;
		}
	}
	
	// reverse -- unsupport
};


#endif // MY_LIST_H_
