#ifndef   MY_SET_H_
#define   MY_SET_H_
/**
	@file myset.h
	@brief set with pool
	@author $Author: dskw $
	@version $Revision: 1.7 $
	@date $Date: 2005/02/18 07:28:43 $
 */

#include <iostream>
#include <new>

using namespace Hotaru;

namespace mysetimp {

enum rb_color { red_ = false, black_ = true };

struct node_base {
	node_base *parent_, *left_, *right_;
	rb_color color_;
	
	static node_base*
	min_node(node_base* n) {
		while(n->left_) n = n->left_;
		return n;
	}
	static node_base*
	max_node(node_base* n) {
		while(n->right_) n = n->right_;
		return n;
	}
	
	static bool not_red(node_base *n) {
		return !n || n->color_ == black_;
	}
};
	
template<typename T_>
struct node : node_base { T_ data_; };

struct iter_base {
	node_base* node_;
	iter_base(){}
	iter_base(node_base* p) : node_(p) {}
	void forward() {
		if(node_->right_) {
			node_ = node_->right_;
			while(node_->left_) node_ = node_->left_;
		} else {
			node_base* p = node_->parent_;
			while(node_ == p->right_) {
				node_ = p;
				p = p->parent_;
			}
			// if !(node_, p) == (head_, root_)
			//  (occurs if 1-node tree)
			if(node_->right_ != p) node_ = p;
		}
	}
	void backward() {
		if(node_->color_ == red_ 
		   && node_->parent_->parent_ == node_) node_ = node_->right_;
		else if(node_->left_) {
			node_base *l = node_->left_;
			while(l->right_) l = l->right_;
			node_ = l;
		} else {
			node_base *p = node_->parent_;
			while(node_ == p->left_) {
				node_ = p;
				p = p->parent_;
			}
			node_ = p;
		}
	}
};

template<typename T_, typename Ref_, typename Ptr_>
struct iter : public iter_base {
	typedef iter<T_, Ref_, Ptr_> Self_;
	typedef node<T_> Node_;
	
	iter(Node_* p) : iter_base(p) {}
	iter() {}
	iter(iter const& i) : iter_base(i.node_) {}
	
	Ref_ operator*() const { return static_cast<Node_*>(node_)->data_; }
	Ptr_ operator->() const { return &(operator*()); }
	Self_& operator++() { forward(); return *this; }
	Self_& operator--() { backward(); return *this; }
	Self_ operator++(int) { Self_ tmp = *this; forward(); return tmp; }
	Self_ operator--(int) { Self_ tmp = *this; backward(); return tmp; }
};

template<typename T_, typename Ref_, typename Ptr_>
inline bool
operator==(iter<T_, Ref_, Ptr_>const& i1, iter<T_, Ref_, Ptr_>const& i2)
	{ return i1.node_ == i2.node_; }
#ifndef _MSC_VER
template<typename T_>
inline bool
operator==(iter<T_,T_&,T_*>const& i1, iter<T_, T_ const&, T_ const*>const& i2)
	{ return i1.node_ == i2.node_; }
template<typename T_>
inline bool
operator==(iter<T_, T_ const&, T_ const*>const& i1, iter<T_,T_&,T_*>const& i2)
	{ return i1.node_ == i2.node_; }
#endif
template<typename T_, typename Ref_, typename Ptr_>
inline bool
operator!=(iter<T_, Ref_, Ptr_>const& i1, iter<T_, Ref_, Ptr_>const& i2)
	{ return i1.node_ != i2.node_; }
#ifndef _MSC_VER
template<typename T_>
inline bool
operator!=(iter<T_,T_&,T_*>const& i1, iter<T_, T_ const&, T_ const*>const& i2)
	{ return i1.node_ != i2.node_; }
template<typename T_>
inline bool
operator!=(iter<T_, T_ const&, T_ const*>const& i1, iter<T_,T_&,T_*>const& i2)
	{ return i1.node_ != i2.node_; }
#endif

inline void
rb_rotL(node_base* n, node_base*& root) {
	node_base *r = n->right_;
	n->right_ = r->left_;
	if(r->left_) r->left_->parent_ = n;
	r->parent_ = n->parent_;
	
	if(n == root) root = r;
	else if(n == n->parent_->left_) n->parent_->left_ = r;
	else n->parent_->right_ = r;
	r->left_ = n;
	n->parent_ = r;
}

inline void
rb_rotR(node_base* n, node_base*& root) {
	node_base *l = n->left_;
	n->left_ = l->right_;
	if(l->right_) l->right_->parent_ = n;
	l->parent_ = n->parent_;
	
	if(n == root) root = l;
	else if(n == n->parent_->right_) n->parent_->right_ = l;
	else n->parent_->left_ = l;
	l->right_ = n;
	n->parent_ = l;
}

inline void
rb_rebalance(node_base* n, node_base*& root) {
	n->color_ = red_;
	while(n != root && n->parent_->color_ == red_) {
		if(n->parent_ == n->parent_->parent_->left_) {
			node_base *r = n->parent_->parent_->right_;
			if(r && r->color_ == red_) {
				n->parent_->color_ = black_;
				r->color_ = black_;
				n->parent_->parent_->color_ = red_;
				n = n->parent_->parent_;
			} else {
				if(n == n->parent_->right_) {
					n = n->parent_;
					rb_rotL(n, root);
				}
				n->parent_->color_ = black_;
				n->parent_->parent_->color_ = red_;
				rb_rotR(n->parent_->parent_, root);
			}
		} else {
			node_base *l = n->parent_->parent_->left_;
			if(l && l->color_ == red_) {
				n->parent_->color_ = black_;
				l->color_ = black_;
				n->parent_->parent_->color_ = red_;
				n = n->parent_->parent_;
			} else {
				if(n == n->parent_->left_) {
					n = n->parent_;
					rb_rotR(n, root);
				}
				n->parent_->color_ = black_;
				n->parent_->parent_->color_ = red_;
				rb_rotL(n->parent_->parent_, root);
			}
		}
	}
	root->color_ = black_;
}

inline node_base*
rb_rebalance_for_erase(node_base* n,
	node_base*& root, node_base*& lm, node_base*& rm) {
	node_base *ret = n;
	node_base *x = 0;
	node_base *p = 0;
	
	if(ret->left_) {
		if(ret->right_) {
			ret = ret->right_;
			while(ret->left_) ret = ret->left_;
			x = ret->right_;
		} else x = ret->left_;
	} else x = ret->right_;
	
	if(ret == n) {
		p = ret->parent_;
		if(x) x->parent_ = ret->parent_;
		if(root == n) root = x;
		else if(n->parent_->left_ == n) n->parent_->left_ = x;
		else n->parent_->right_ = x;
		if(lm == n) {
			if(n->right_) lm = node_base::min_node(x);
			else lm = n->parent_;
		}
		if(rm == n) {
			if(n->left_) rm = node_base::max_node(x);
			else rm = n->parent_;
		}
	} else {
		n->left_->parent_ = ret;
		ret->left_ = n->left_;
		if(ret == n->right_) p = ret;
		else {
			p = ret->parent_;
			if(x) x->parent_ = ret->parent_;
			ret->parent_->left_ = x;
			ret->right_ = n->right_;
			n->right_->parent_ = ret;
		}
		if(root == n) root = ret;
		else if(n->parent_->left_ == n) n->parent_->left_ = ret;
		else n->parent_->right_ = ret;
		ret->parent_ = n->parent_;
		std::swap(ret->color_, n->color_);
		ret = n;
	}
	
	if(ret->color_ == black_) {
		while(x != root && node_base::not_red(x)) {
			if(x == p->left_) {
				node_base *w = p->right_;
				if(w->color_ == red_) {
					w->color_ = black_;
					p->color_ = red_;
					rb_rotL(p, root);
					w = p->right_;
				}
				if(node_base::not_red(w->left_) &&
				   node_base::not_red(w->right_)) {
					w->color_ = red_;
					x = p;
					p = p->parent_;
				} else {
					if(node_base::not_red(w->right_)) {
						w->left_->color_ = black_;
						w->color_ = red_;
						rb_rotR(w, root);
						w = p->right_;
					}
					w->color_ = p->color_;
					p->color_ = black_;
					if(w->right_) w->right_->color_ = black_;
					rb_rotL(p, root);
					break;
				}
			} else {
				node_base *w = p->left_;
				if(w->color_ == red_) {
					w->color_ = black_;
					p->color_ = red_;
					rb_rotR(p, root);
					w = p->left_;
				}
				if(node_base::not_red(w->right_) &&
				   node_base::not_red(w->left_)) {
					w->color_ = red_;
					x = p;
					p = p->parent_;
				} else {
					if(node_base::not_red(w->left_)) {
						w->right_->color_ = black_;
						w->color_ = red_;
						rb_rotL(w, root);
						w = p->left_;
					}
					w->color_ = p->color_;
					p->color_ = black_;
					if(w->left_) w->left_->color_ = black_;
					rb_rotR(p, root);
					break;
				}
			}
		}
		if(x) x->color_ = black_;
	}
	
	return ret;
}


template<typename T_, int num_, size_t size_ = sizeof(node<T_>)>
struct mypool {
	typedef union tagE { char mem_[size_]; union tagE* ptr_; } Elem_;
	std::vector<Elem_*> vpool_;
	Elem_ *first_;
	mypool() : first_(0) {
		vpool_.reserve(100);
	}
	~mypool() {
		for(size_t i = 0; i < vpool_.size(); ++i) delete vpool_[i];
	}
	Elem_*
	alloc_new_pool() {
std::cerr << "num pools[" << num_ << ", "
	<< sizeof(Elem_) << "]: " << vpool_.size()+1 << std::endl;
		Elem_ *tmp = new Elem_[num_];
		vpool_.push_back(tmp);
		for(int i = 1; i < num_; ++i) tmp[i-1].ptr_ = &tmp[i];
		tmp[num_ - 1].ptr_ = 0;
		return tmp;
	}
	
	void* alloc() {
		if(!first_) first_ = alloc_new_pool();
		Elem_* ret = first_;
		first_ = first_->ptr_;
		return ret;
	}
	void dealloc(void *p) {
		Elem_ *pp = (Elem_*)p;
		pp->ptr_ = first_;
		first_ = pp;
	}
};


template<typename T_,  int num_, size_t size_ = sizeof(node<T_>)>
struct tree_base {
	typedef node<T_> Node_;
	Node_ *head_;
	
	tree_base() : head_(get_node()) {}
	~tree_base() { put_node(head_); }
	
	static mypool<T_, num_, size_> pool_;
	static Node_* get_node() { return (Node_*)pool_.alloc(); }
	static void put_node(Node_* p) { pool_.dealloc(p); }
};

template<typename K_, typename T_, typename C_, int num_>
struct tree : tree_base<T_, num_> {
	typedef node<T_> Node_;
	typedef iter<T_, T_ const&, T_ const*> iterator;
	typedef std::pair<iterator, bool> Pair_;
	
	size_t count_;
	C_ comp_;
	
	Node_*& root_() const { return (Node_*&)head_->parent_; }
	Node_*& lm_() const { return (Node_*&)head_->left_; }
	Node_*& rm_() const { return (Node_*&)head_->right_; }
	
	Node_*
	copy_(node_base* r, Node_* p) {
		Node_* t = clone_(r);
		t->parent_ = p;
		
		try {
			if(r->right_) t->right_ = copy_(r->right_, t);
			p = t;
			r = r->left_;
			while(r) {
				Node_* l = clone_(r);
				p->left_ = l;
				l->parent_ = p;
				if(r->right_) l->right_ = copy_(r->right_, l);
				p = l;
				r = r->left_;
			}
		} catch(...) {
			erase_(t);
			throw;
		}
		return t;
	}
	void
	erase_(node_base* r) {
		while(r) {
			erase_(r->right_);
			node_base* l = r->left_;
			destroy_(r);
			r = l;
		}
	}
	
	Node_*
	create_(T_ const& r) {
		Node_* tmp = get_node();
		try {
			new (static_cast<void*>(&tmp->data_)) T_(r);
		} catch(...) {
			put_node(tmp);
			throw;
		}
		return tmp;
	}
	
	Node_*
	clone_(node_base* r) {
		Node_* tmp = create_(static_cast<Node_*>(r)->data_);
		tmp->color_ = r->color_;
		tmp->left_ = 0;
		tmp->right_ = 0;
		return tmp;
	}
	void
	destroy_(node_base* r) {
		static_cast<Node_*>(r)->data_.~T_();
		put_node(static_cast<Node_*>(r));
	}
	
	void
	init_() {
		color_(head_) = red_;
		root_() = 0;
		lm_() = head_;
		rm_() = head_;
	}
	iterator
	insert_(Node_* x, Node_* y, T_ const& v) {
		Node_* z = create_(v);
		if(y == head_ || x || comp_(v, y->data_)) {
			y->left_ = z;
			if(y == head_) {
				root_() = z;
				rm_() = z;
			} else if(y == lm_()) lm_() = z;
		} else {
			y->right_ = z;
			if(y == rm_()) rm_() = z;
		}
		z->parent_ = y;
		z->left_ = 0;
		z->right_ = 0;
		rb_rebalance(z, head_->parent_);
		++count_;
		return iterator(z);
	}
	
	static Node_* min_(Node_*& r) { return (Node_*)node_base::min_node(r); }
	static Node_* max_(Node_*& r) { return (Node_*)node_base::max_node(r); }
	static rb_color& color_(Node_* r) { return r->color_; }
	
	// ctor/dtor
	tree() : tree_base<T_, num_>(), count_(0), comp_() { init_(); }
	explicit tree(C_ const& comp)
		: tree_base<T_, num_>(), count_(0), comp_(comp) { init_(); }
	explicit tree(tree<K_, T_, C_, num_>const& r)
		: tree_base<T_, num_>(), count_(0), comp_(r.comp_) {
		if(r.root_()) {
			color_(head_) = red_; // why need and not at op=?
			root_() = copy_(r.root_(), head_);
			lm_() = min_(root_());
			rm_() = max_(root_());
		} else init_();
		count_ = r.count_;
	}
	~tree() { clear(); }
	
	// begin/end
	iterator begin() const { return lm_(); }
	iterator end() const { return head_; }
	
	// empty/size
	bool empty() const { return count_ == 0; }
	size_t size() const { return count_; }
	
	// swap
	void
	swap(tree<K_, T_, C_, num_>& r) {
		std::swap(head_, r.head_);
		std::swap(count_, r.count_);
		std::swap(comp_, r.comp_);
	}
	
	// insert/erase
	Pair_
	insert_unique(T_ const& r) {
		Node_ *h = head_, *n = root_();
		bool comp = true;
		while(n) {
			h = n;
			comp = comp_(r, n->data_);
			n = static_cast<Node_*>(comp? n->left_: n->right_);
		}
		iterator s = iterator(h);
		if(comp) {
			if(s == begin()) return Pair_(insert_(n, h, r), true);
			else --s;
		}
		if(comp_(static_cast<Node_*>(s.node_)->data_, r)) {
			return Pair_(insert_(n, h, r), true);
		}
		return Pair_(s, false);
	}
	
	iterator
	insert_unique(iterator i, T_ const& r) {
		if(i.node_ == head_->left_) { // begin()
			return insert_(static_cast<Node_*>(i.node_),
				static_cast<Node_*>(i.node_), r);
		} else if(i.node_ == head_) { // end()
			return insert_(0, rm_(), r);
		}
		iterator b = i; --b;
		if(b.node_->right_) {
			return insert_(static_cast<Node_*>(i.node_),
				static_cast<Node_*>(i.node_), r);
		}
		return insert_(0, static_cast<Node_*>(b.node_), r);
	}
	
	void
	erase(iterator r) {
		destroy_(rb_rebalance_for_erase(
			r.node_, head_->parent_, head_->left_, head_->right_));
		--count_;
	}
	
	// clear
	void
	clear() {
		if(!count_) return;
		erase_(root_());
		lm_() = head_;
		root_() = 0;
		rm_() = head_;
		count_ = 0;
	}
	
	// set ops
	
	// op=
	tree<K_, T_, C_, num_>&
	operator=(tree<K_, T_, C_, num_>const& r) {
		if(this != &r) {
			clear();
			comp_ = r.comp_;
			if(r.root_()) {
				root_() = copy_(r.root_(), head_);
				lm_() = min_(root_());
				rm_() = max_(root_());
				count_ = r.count_;
			}
		}
		return *this;
	}
};



}

template<typename T_, typename C_, int num_ = POOL_NODE_NUM>
struct myset {
	typedef mysetimp::iter<T_, T_ const&, T_ const*> const_iterator;
	typedef mysetimp::iter<T_, T_ const&, T_ const*> iterator;
	typedef mysetimp::tree<T_, T_, C_, num_> Rep_;
	
	Rep_ tree_;
	
	// ctor/op=/dtor -- with no allocator
	// unsupport : 
	//   ctor(InIter f, InIter l, C_ const& comp_)
	myset() : tree_(C_()) {}
	explicit myset(C_ const& comp) : tree_(comp_) {}
	myset(myset const& r) : tree_(r.tree_) {}
	myset<T_, C_>& operator=(myset<T_, C_>const& r) {
		tree_ = r.tree_;
		return *this;
	}
	
	// get_allocator -- unsupport
	// get compare -- unsupport
	
	// begin/end -- reverse:unsupport
        const_iterator begin() const { return tree_.begin(); }
        const_iterator end() const { return tree_.end(); }
	
	// empty/size -- resize/max_size:unsupport
	bool empty() const { return tree_.empty(); }
	size_t size() const { return tree_.size(); }
	
	// insert
	std::pair<const_iterator, bool> insert(T_ const& r) {
		return tree_.insert_unique(r);
	}
	const_iterator insert(const_iterator i, T_ const& r) {
		return tree_.insert_unique(i, r);
	}
	
	// erase
	void erase(const_iterator r) { tree_.erase(r); }
	
	// swap
	void swap(myset& r) { tree_.swap(r.tree_); }
	
	// clear
	void clear() { tree_.clear(); }
	
	// set ops
	// unsupported
};


#endif // MY_LIST_H_
