/**
	ʰץ󥽡
	@brief
	@author $Author: dskw $
	@version $Revision: 1.7 $
	@date $Date: 2005/02/24 04:17:58 $
*/

#include "hotaru_f4.h"
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <map>

using namespace std;

namespace Hotaru {
namespace ParaF4 {

const char* getString() { return "ParaF4"; }

static int nFirst, nLast;
static bool isFree[MAX_PE_ID];
static bool reassign[MAX_PE_ID];
static struct stat sb;
static char fname[32];
static map<F2poly const*, int> ip;

void
InitChildren() {
	int nProc = 0;
	
	::memset(isFree, 0, sizeof isFree);
	for(int n = nFirst; n < nLast; ++n) {
		sprintf(fname, "F%d", n);
		isFree[n] = !::stat(fname, &sb);
		if(isFree[n]) unlink(fname), ++nProc;
		reassign[n] = true;
	}
	if(!nProc) {
		std::cerr << "ERROR!! No processor elements available"
			<< std::endl;
		exit(1);
	}
}

static void
PutG(F2poly::Set const& G, int n) {
	::sprintf(fname, "G%d", n);
	ofstream outG(fname, ios::trunc|ios::out);
	for(F2poly::Set::const_iterator iG = G.begin(); iG != G.end(); ++iG) {
		outG << *iG << endl;
	}
	outG.close();
}

static void
PutP(F4::PairList const& Pd, int n) {
	::sprintf(fname, "_P%d", n);
	ofstream outP(fname, ios::trunc|ios::out);
	F4::PairList::const_iterator iPd = Pd.begin();
	for(; iPd != Pd.end(); ++iPd) {
		F2poly const* f1 = &iPd->f1_;
		F2poly const* f2 = &iPd->f2_;
		if(ip.find(f1) == ip.end() ||
		   ip.find(f2) == ip.end()) {
			std::cerr << "warning: pair is lost" << std::endl;
			continue;
		}
		int i1 = ip[f1], i2 = ip[f2];
		if(i1 == i2) i2 = ~iPd->var_;
		outP << i1 << ',' << i2 << endl;
	}
	outP.close();
	::rename(fname, fname+1);
}


void
MakeGbaseBy(F2poly::Set& G) {
	F2poly::Set F;
	F4::PairList P, Pd;
	F4::UpdateGM U;
	F4::SelectMinDegLCM S;
	//F4::SelectMinDegLCM_LT S(1000);
	
	
	F.swap(G);
	F4::Resolve(F);
	F4::Reduce(F);
	U(G, P, F);
	F2poly::Set::iterator iG = G.begin();
	for(int id = 0; iG != G.end(); ++iG, ++id) ip[&*iG] = id;
	
	int nChild = 0;
	while(nChild > 0 || P.size() > 0) {
		bool waiting = true;
		
		// assign Pairs to children
		for(int n = nFirst; n < nLast; ++n) {
			if(!P.size()) break;
			if(!isFree[n]) continue;
			
			isFree[n] = false;
			++nChild;
			S(P, Pd);
			PutG(G, n);
			PutP(Pd, n);
cerr << "Pair(size = " << Pd.size() << ") assigned: PE#"
	<< n << " running #of children: " << nChild << endl;
			Pd.clear();
			::rename(fname, fname+1);
			waiting = false;
		}
		
		// get result from children
		for(int n = nFirst; n < nLast; ++n) {
			if(isFree[n]) continue;
			sprintf(fname, "F%d", n);
			if(::stat(fname, &sb)) continue;
			
			isFree[n] = true;
			--nChild;
			
			F2poly::Set F;
			string s;
			ifstream inF(fname);
			while(getline(inF, s)) F.push_back(F2poly(s.c_str()));
			inF.close();
			unlink(fname);
cerr << "Polynomials(size = " << F.size() << ") got from: PE#"
	<< n << " running #of children: " << nChild << endl;
			int nL = G.size();
			
			if(F.size() && reassign[n]) {
				sprintf(fname, "G%d", n);
				ifstream inG(fname);
				int nG = 0;
				while(getline(inG, s)) ++nG;
				if(nG != nL) {
					++nChild;
					isFree[n] = false;
cerr << "Reassign to PE#" << n << " running #of children: " << nChild << endl;
					PutG(G, n);
					PutP(F4::PairList(), n);
					reassign[n] = false;
					continue;
				}
			
				U(G, P, F);
				F2poly::Set::iterator iG = G.begin();
				std::advance(iG, nL);
				for(; iG != G.end(); ++iG) ip[&*iG] = nL++;
cerr << "#of G = " << G.size() << ",  #of P = " << P.size() << endl;
			}
			waiting = false;
			reassign[n] = true;
			//reduced = false;
			n = nFirst - 1;
		}
		if(waiting) ::usleep(10000);
	}
	
	F4::Reduce(G);
}

} //ParaF4
} //Hotaru
