/**
	ʰץ󥽡
	@brief
	@author $Author: dskw $
	@version $Revision: 1.4 $
	@date $Date: 2005/02/23 11:11:15 $
*/

#include "hotaru_f4.h"
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <map>

using namespace std;

namespace Hotaru {
namespace ParaHFE {

const char* getString() { return "ParaHFE"; }

static void
join(int n, F2poly::Set& G1, F2poly::Set& G2, F2poly::Set& G) {
	F2term t; t.setVar(n-1);
	F2poly::Set::iterator iG;
	
std::cout << "G1(to be mul by " << t << "+1) = {" << std::endl;
	// G1: when var(n-1) == 0
	for(iG = G1.begin(); iG != G1.end(); ++iG) {
std::cout << "  " << *iG << "," << std::endl;
		F2poly f(*iG); f *= t;
		*iG += f;
	}
std::cout << "},\nG2(to be mul by " << t << ") = {" << std::endl;
	// G2: when var(n-1) == 1
	for(iG = G2.begin(); iG != G2.end(); ++iG) {
std::cout << "  " << *iG << "," << std::endl;
		*iG *= t;
	}
	G.splice(G.end(), G1, G1.begin(), G1.end());
	G.splice(G.end(), G2, G2.begin(), G2.end());
	F4::Reduce(G);
std::cout << "},\nG(joined) = {" << std::endl;
for(iG = G.begin(); iG != G.end(); ++iG) {
std::cout << "  " << *iG << "," << std::endl;
}
std::cout << "}\n" << std::endl;
}

template<int n>
struct SubProblem : SubProblem<n-1> {
	enum { num_ = 1 << n };
	bool toUp_[num_];
	F2poly::Set G_[num_];
	
	SubProblem() { ::memset(toUp_, 0, sizeof toUp_); }
	bool doUp() {
		int dif = 1 << n-1;
		for(int i = 0; i < dif; ++i) if(toUp_[i] && toUp_[i+dif]) {
std::cout << "@level" << n << ", joining " << i << " and " << i+dif
	<< std::endl;
			join(n, G_[i], G_[i+dif], SubProblem<n-1>::G_[i]);
			toUp_[i] = toUp_[i+dif] = false;
			SubProblem<n-1>::toUp_[i] = true;
		}
		return SubProblem<n-1>::doUp();
	}
};

template<>
struct SubProblem<1> {
	enum { num_ = 2 };
	bool toUp_[num_];
	F2poly::Set G_[num_];
	F2poly::Set G0_;
	
	SubProblem() { toUp_[0] = toUp_[1] = false; }
	bool doUp() {
		if(!(toUp_[0] && toUp_[1])) return false;
std::cout << "@level1, joining 0 and 1" << std::endl;
		join(1, G_[0], G_[1], G0_);
		return true;
	}
	
	F2poly::Set const& G() const { return G0_; }
};

static int nLevel;
static struct stat sb;
static char fname[32];

static void
PutF(F2poly::Set const& F, int n) {
	::sprintf(fname, "F%d", n);
	ofstream outF(fname, ios::trunc|ios::out);
	for(F2poly::Set::const_iterator iF = F.begin(); iF != F.end(); ++iF) {
		outF << *iF << endl;
	}
	outF.close();
}

static void
DelG(int n) {
	::sprintf(fname, "G%d", n);
	::unlink(fname);
}

template<int n>
void
Dispatch(F2poly::Set& G) {
	SubProblem<n> s;
	
	// dispatch subproblem
	F2poly f[2][n];
	for(int i = 0; i < n; ++i) {
		F2term t; t.setVar(i);
		f[0][i] = t; f[1][i] = t; f[1][i] += F2term();
	}
	for(int i = 0; i < (1<<n); ++i) {
		F2poly::Set F(G);
		for(int j = 0; j < n; ++j) F.push_back(f[!!(i&(1<<j))][j]);
		DelG(i);
		PutF(F, i);
	}

	// wait for answer
	bool got[1<<n];
	::memset(&got, 0, sizeof got);
	for(;;) {
		for(int i = 0; i < (1<<n); ++i) {
			if(got[i]) continue;
			sprintf(fname, "G%d", i);
			if(::stat(fname, &sb)) continue;
			
			got[i] = true;
			
			string st;
			ifstream inG(fname);
			while(getline(inG, st)) {
				s.G_[i].push_back(F2poly(st.c_str()));
			}
			inG.close();
			unlink(fname);
			s.toUp_[i] = true;
			i = -1;
		}
		if(s.doUp()) {
			G.clear();
			G = s.G();
			break;
		}
	}
}

void
MakeGbaseBy(F2poly::Set& G) {
	if(nLevel > 10) {
		std::cerr << "level must be less than 10" << std::endl;
	}
	switch(nLevel) {
	case 10: Dispatch<10>(G); break;
	case  9: Dispatch<9>(G);  break;
	case  8: Dispatch<8>(G);  break;
	case  7: Dispatch<7>(G);  break;
	case  6: Dispatch<6>(G);  break;
	case  5: Dispatch<5>(G);  break;
	case  4: Dispatch<4>(G);  break;
	case  3: Dispatch<3>(G);  break;
	case  2: Dispatch<2>(G);  break;
	case  1: Dispatch<1>(G);  break;
	}
}
	
} //ParaHFE
} //Hotaru
